#!/usr/bin/perl
#
# initdb.pl      Norbert Haider, University of Vienna, 2009-2021
#                norbert.haider@univie.ac.at
#
# This script is part of the MolDB6 package. Last change: 2021-02-27
#
# This script must be run before anything else in order to (re-)create 
# the MolDB6 database and to grant appropriate rights to the proxy users
#
# The "mysql" command must be in your search path.
#
# Starting with MolDB5R.03, this script also initializes the fragment
# dictionary, again by invoking the "mysql" command-line client (in
# previous versions, this was done by calling the initfpdef.pl script
# which requires the "DBI" Perl database driver package to be installed).
#
# Usage: 
# 
# perl initdb.pl
#   or
# perl initdb.pl --dryrun
#

# defaults
$oldmysql      = 0;  # 0 for MySQL >= 5.7, 1 for older versions
$fpdict_mode   = 1;  # 1 = auto adjust, 2 = force 64 bit, 3 = force 32 bit
$fpfile        = "fp01.sdf";


$os  = "";
$win = 0;
$os = uc($ENV{OS});
if ($os eq "") { $os = uc($ENV{OSTYPE}); }
if (index($os,"WINDOWS")>=0) {
  #print "using Windows workaround\n";
  $win = 1;
}

$dry = 0;
if ($#ARGV < 0) {
  $userinput = "";
} else {
  $userinput = $ARGV[0];
  if ($userinput == "--dryrun") { $dry = 1; }
}


# check for mysql command-line client
$return = `mysql -V`;
if (index($return,"mysql") < 0) {
  print STDERR "ERROR: could not find 'mysql' client program, make sure it is installed and in\nyour search path!";
  exit;
}


$configfile="./moldb6.conf";      # read all settings from this file
$return     = do $configfile;

# settings that override the configuration file
$tweakmolfiles = "y";

if (!defined $return) { die("ERROR: cannot read configuration file $configfile!"); }	
if ($database eq "") { die("ERROR: no database specified"); }
if ($hostname eq "") { $hostname = "localhost"; }
if ($clientname eq "") { $clientname = "localhost"; }
if ($mysql_admin eq "") { $mysql_admin = "root"; }
if ($metatable eq "") { $metatable = "moldb_meta"; }
if ($rw_user eq "") { die("ERROR: no rw_user specified"); }
if ($rw_password eq "") { die("ERROR: no rw_password specified"); }
if ($ro_user eq "") { die("ERROR: no ro_user specified"); }
if ($ro_password eq "") { die("ERROR: no ro_password specified"); }
if ($charset eq "") { $charset = "latin1"; }

print "\nCreating a MolDB6 database with the following configuration:\n";
print "database:    $database\n";
print "hostname:    $hostname\n";
print "clientname:  $clientname\n";
print "mysql_admin: $mysql_admin\n";
print "rw_user:     $rw_user\n";
print "ro_user:     $ro_user\n";
print "drop_db:     $drop_db\n";
print "charset:     $charset\n\n";

if ($dry eq 0) {
  print "do you want to continue (y/n)? ";
  chomp($word = <STDIN>);
  if (lc($word) eq 'y') {
      print "OK\n";
  } else {
    print "aborting\n";
    exit;
  }
}

if ($charset eq "latin1") {
  $mysql_charset = "latin1";
  $mysql_collation = "latin1_swedish_ci";
}
if ($charset eq "latin2") {
  $mysql_charset = "latin2";
  $mysql_collation = "latin2_general_ci";
}
if ($charset eq "utf8") {
  $mysql_charset = "utf8";
  $mysql_collation = "utf8_unicode_ci";
}



$cmd = "";

if ($drop_db eq "y") {
  print "\nATTENTION! An already existing database $database will be erased!\n";
  $cmd ="DROP DATABASE IF EXISTS $database; ";
}

if ($dry eq 0) {
  print "\nYou will be prompted for the password of the MySQL administrator ($mysql_admin)\n";
  print "If you want to cancel, just hit <Return>\n";
}

$cmd  .= "CREATE DATABASE IF NOT EXISTS $database; \n";

if ($oldmysql eq 0) {
  $cmd .= "DROP USER IF EXISTS '" . $rw_user . "'\@'" . $clientname . "';\n";
  $cmd .= "DROP USER IF EXISTS '" . $ro_user . "'\@'" . $clientname . "';\n";   
  $cmd .= "CREATE USER IF NOT EXISTS '" . $rw_user . "'\@'" . $clientname . "' IDENTIFIED BY '" . $rw_password . "';\n";
  $cmd .= "CREATE USER IF NOT EXISTS '" . $ro_user . "'\@'" . $clientname ."' IDENTIFIED BY '" . $ro_password ."';\n";
  $cmd .= "GRANT ALL on $database.* TO '" . $rw_user . "'\@'" . $clientname . "';\n";
  $cmd .= "GRANT FILE ON *.* TO '" . $rw_user . "'\@'" . $clientname . "';\n";
  $cmd .= "GRANT SELECT ON $database.* TO '" . $ro_user . "'\@'" . $clientname ."';\n";
} else {
  $cmd .= "GRANT ALL on $database.* TO '" . $rw_user . "'\@'" . $clientname . "' IDENTIFIED BY '" . $rw_password . "';\n";
  $cmd .= "GRANT FILE ON *.* TO '" . $rw_user . "'\@'" . $clientname . "';\n";
  $cmd .= "GRANT SELECT ON $database.* TO '" . $ro_user . "'\@'" . $clientname ."' IDENTIFIED BY '" . $ro_password ."';\n";
}

$cmd .= "USE $database; DROP TABLE IF EXISTS $metatable;
CREATE TABLE $metatable (
  db_id INT(11) NOT NULL,
  type TINYINT(4) UNSIGNED NOT NULL DEFAULT '1' COMMENT '1 = substance, 2 = reaction, 3 = combined',
  access TINYINT(4) UNSIGNED NOT NULL DEFAULT '1' COMMENT '0=hidden, 1=read-only, 2=add/update, 3=full access',
  name VARBINARY(255) NOT NULL,
  description TINYTEXT CHARACTER SET $mysql_charset COLLATE $mysql_collation NOT NULL,
  usemem ENUM('T','F') NOT NULL DEFAULT 'F',
  memstatus TINYINT(4) UNSIGNED NOT NULL DEFAULT '0',
  digits TINYINT(4) UNSIGNED NOT NULL DEFAULT '8',
  subdirdigits TINYINT(4) UNSIGNED NOT NULL DEFAULT '4',
  trustedIP VARBINARY(255) NOT NULL,
  flags int(11) UNSIGNED not NULL DEFAULT '0',
  PRIMARY KEY  (db_id)
) ENGINE=MyISAM DEFAULT CHARSET=binary COMMENT='meta information about MolDB6 data collections';";

if ($dry eq 1) {
  print "\n------------------------------ dryrun mode ------------------------------\n";
  print "The following SQL statements would have been executed as MySQL administrator:\n\n";
  print "$cmd\n";
} else {

  if ($win eq 1) {
    open(SQL,">initdb.sql");
    print SQL "$cmd";
    close(SQL);
    $result=`mysql -h $hostname -u $mysql_admin -p < initdb.sql 2>&1`;
    unlink("initdb.sql");
  } else {
    $result=`echo \"$cmd\" | mysql -h $hostname -u $mysql_admin -p 2>&1`;
  }
  
  if (index($result,"ERROR") >= 0) {
    die("creating table $metatable in database $database failed\n  $result");
  } else {
    print "setup of database $database OK\n";
  }

}

# now set up the fragment dictionary


open (FPF, "<$fpfile") || die ("cannot open fingerprint file $fpfile!");

$user     = $rw_user;    # from configuration file
$password = $rw_password;

$cmd  = "USE $database;\n";
$cmd .= "DROP TABLE IF EXISTS $fpdeftable;\n";
$cmd .= "CREATE TABLE $fpdeftable (fp_id INT(11), fpdef MEDIUMBLOB, fptype SMALLINT) ENGINE=MyISAM;\n";

$fpstruc = "";
$n_dict  = 0;
$n_mol   = 0;
$molbuf  = "";
$fptype  = 1;  # 1 = 64-bit, 2 = 32-bit
if ($fpdict_mode < 3) { $max_bs = 62; } else { $max_bs = 31; }

while ($line = <FPF>) {
  if ((substr($line,0,4) eq '$$$$') || eof(FPF)) {
    if (length($molbuf)>20) {  
      if (length($fpstruc) > 20) { $fpstruc .= '$$$$ ' . "\r\n"; }
      if (index($molbuf,"M  END")<0) { $molbuf .= "M  END\r\n"; }
      $fpstruc .= $molbuf;
      $molbuf = "";
      $n_mol++;
    }
    if (($n_mol % $max_bs == 0) || eof(FPF)) {
      # insert dictionary chunk into fpdef table
      if ($n_mol > 31) {
      	$fptype = 1;
      } else {
        if ($fpdict_mode == 2) { $fptype = 1; } else { $fptype = 2; }
      }
      $n_dict++;
      #print "$fpstruc\n";
      #print "chunk $n_dict contains $n_mol structures (type $fptype)\n";
      $cmd .= "INSERT INTO $fpdeftable VALUES ($n_dict, \"$fpstruc\", $fptype );\n";      
      $fpstruc = "";
      $molbuf = "";
      $n_mol = 0;
    }
  } else {
    $molbuf = $molbuf . $line;
  }
}
close(FPF);

if ($dry eq 1) {
  print "$cmd\n";
} else {

  if ($win eq 1) {
    open(SQL,">initfpdef.sql");
    print SQL "$cmd";
    close(SQL);
    $result=`mysql -h $hostname -u $rw_user --password=$rw_password < initfpdef.sql 2>&1`;
    unlink("initfpdef.sql");
  } else {
    $cmd =~ s/\$/\\\$/g;
    $cmd =~ s/\"/\\\"/g;
    $result=`echo \"$cmd\" | mysql -h $hostname -u $rw_user --password=$rw_password 2>&1`;
  }
  
  if (index($result,"ERROR") >= 0) {
    die("setting up fragment dictionary in database $database failed\n  $result");
  } else {
    print "setup of fragment dictionary OK\n";
  }

}