#!/usr/bin/perl
#
# mk3d.pl           Norbert Haider, University of Vienna, 2016-2018
#                   norbert.haider@univie.ac.at
#
# This script is an inofficial part of the MolDB6 package. Last change: 2018-01-05
#
# Example script which (re-)generates the 3D molfiles for 3D depiction
# of the molecules in the MolDB6 database by energy minimization, using
# two different force fields as implemented in the obminimize command.

# Requirements: OpenBabel must be installed.

# general settings:

$dblist  = "1,2"; # a comma-separated list of the data collection(s) for which 3D 
                # structures are to be generated (e.g., "1,2,4,7")
$basedir = "/var/www/htdocs/moldb6/struc3d"; # the 3D diercotory (full path)
$baseURLdir = "struc3d";        # the 3D directory (only URL-relevant path)
$struc3dfield = "struc3d";  # the MySQL field (column) in which the URL is stored
$OBMINIMIZE = "/usr/bin/obminimize";

use DBI();
use File::Temp qw/ tempfile tempdir /; 
use File::Copy;
use File::Spec;

$level      = 3;   # 1 = process all entries with empty struc3d field (i.e., all new molecules)
                   # 2 = like 1, but also check for missing 3D molfiles
                   # 3 = like 2, but also check for orphan 3D molfiles and delete them

$verbose    = 2;   # 0 = silent operation, 
                   # 1 = report each data collection + principal action
                   # 2 = report each molecule

$reallydelete = 1; # 0 = only rename files which are no longer needed
                   # 1 = really delete these files
                   

$configfile = "../moldb6.conf";
#$configfile = "/data/moldb/moldb6/moldb6.conf";  # better use absolute path when
                                                   # running this script via cron

if ($#ARGV >= 0) {
  for ($a = 0; $a <= $#ARGV; $a++) {
    $aline = $ARGV[$a];
    if (($aline eq '-h') || ($aline eq '-?') || ($aline eq '--help')) {
      print "Usag: mk3d.pl [--level=<n>]    (with <n> as 1, 2, 3, or full)\n";
      exit;
    }
    if (index($aline,"--level=") >= 0) {
      $levelval = getval($aline);
      if ($levelval eq "1") { $level = 1; }
      if ($levelval eq "2") { $level = 2; }
      if ($levelval eq "3") { $level = 3; }
      if ($levelval eq "full") { $level = 3; }
      
    }
  }
}


$ostype = getostype();

$return     = do $configfile;
if (!defined $return) {
  die("ERROR: cannot read configuration file $configfile!\n");
}	

$user     = $rw_user;    # from configuration file
$password = $rw_password;

$mysql_charset   = "latin1";          # default
$mysql_collation = "latin1_swedish_ci";

if ($charset eq "latin2") {           # from configuration file
  $mysql_charset   = "latin2";
  $mysql_collation = "latin2_general_ci";
}
if ($charset eq "utf8") {
  $mysql_charset   = "utf8";
  $mysql_collation = "utf8_unicode_ci";
}

if ((defined $basedir) && ($basedir ne "")) {
  if ( ! -d $basedir) {
    mkdir $basedir || die("ERROR: cannot create directory $basedir !");
    if ($verbose > 0) {
      print "created directory: $basedir\n";
    }
  }
  if ( ! -W $basedir) {
    die("ERROR: cannot write to directory $basedir !");
  }
  # check for obminimize
  $return = `$OBMINIMIZE`;
  if (index($return,"Usage:") < 0) {
    die("ERROR: could not find 'obminimize', make sure that OpenBabel is installed");	
  }  
  
} else {
  die("ERROR: 3D directory is not defined in your configuration file!");
}


@dbarr = split(/,/,$dblist);
if ($verbose > 0) {
  print "creating 3D structures for data collection(s):";

  for ($i = 0; $i <= $#dbarr; $i++) {
    $db_num = $dbarr[$i];
    print " $db_num";
  }
  print "\n";
}




$dbh = DBI->connect("DBI:mysql:database=$database;host=$hostname",
                    $user, $password,
                    {'RaiseError' => 1});
$qstr = "SET SESSION sql_mode = '';";
$dbh->do($qstr);

# read moldb_meta table and find out which data collections are to be processed
$ndb = 0;
@db = [""];
#$sth0 = $dbh->prepare("SELECT db_id FROM $metatable WHERE type = 1 ORDER BY db_id");
$sth0 = $dbh->prepare("SELECT db_id, type FROM $metatable WHERE type = 1 ORDER BY db_id");
$sth0->execute();
while ($ref0 = $sth0->fetchrow_hashref()) {
  $db_id   = $ref0->{'db_id'};
  $db_type = $ref0->{'type'};
  for ($i = 0; $i <= $#dbarr; $i++) {
    $db_num = $dbarr[$i];
    if ($db_num == $db_id) {
      #print "Yes, this is a relevant db: $db_id \n";
      $ndb++;
      @db[($ndb-1)] = [ $db_id, $db_type ];
    }
  }
}
$sth0->finish;


#exit;


$badmols = 0;
$counter = 0;
$delcounter = 0;

for ($idb = 0; $idb < $ndb; $idb++) {
  $dbnum  = $db[$idb][0];
  $dbtype = $db[$idb][1];
  $dbprefix = $prefix . "db" . $dbnum . "_";
  if ($dbtype == 1) {
    $idname = "mol_id";
    $structable = $dbprefix . $molstrucsuffix;
    $datatable = $dbprefix . $moldatasuffix;
  }

  # check if this data collection really has a struc3d field
  $found_3dfield = 0;
  $qstr = "DESCRIBE $datatable";
  $sth0 = $dbh->prepare($qstr);
  $sth0->execute();
  while ($ref0    = $sth0->fetchrow_hashref()) {
    $field       = $ref0->{'Field'};
    if ($field eq $struc3dfield) { $found_3dfield = 1; }
  }
  $sth0->finish;
  if ($found_3dfield == 0) {
    if ($verbose > 0) {
      print "data collection $dbnum does not contain a $struc3dfield column - skipping...\n";
    }
    next;  # skip this data collection and check the next one in the list
  }
  
  # get the number of digits and subdir digits for MOL filenames of this data collection
  $sth0 = $dbh->prepare("SELECT digits, subdirdigits FROM $metatable WHERE db_id = $dbnum");
  $sth0->execute();
  while ($ref0    = $sth0->fetchrow_hashref()) {
    $digits       = $ref0->{'digits'};
    $subdirdigits = $ref0->{'subdirdigits'};
  }
  $sth0->finish;
  # check if the number of MOL filename digits is OK
  if (!defined $digits) {
    $digits = 8;
    if ($verbose > 0) {
      print "could not get number of digits, using default ($digits)\n";
    }
  }
  # check if subdirectories are to be used
  if (!defined $subdirdigits) {
    $subdirdigits = 0;
    if ($verbose > 0) {
      print "could not get number of subdir digits, using default ($subdirdigits)\n";
    }
  }

  # get the number of entries to be processed and divide it into smaller chunks  
  $sth0 = $dbh->prepare("SELECT COUNT($idname) AS itemcount FROM $structable ");
  $sth0->execute();
  while ($ref0 = $sth0->fetchrow_hashref()) {
    $itemcount = $ref0->{'itemcount'};
  }
  $sth0->finish();
  if ($verbose > 0) {
    print "number of entries in data collection $dbnum to be processed: $itemcount \n";
  }
  $nchunks = int( (($itemcount + 99) / 100) );
  $li  = 0;
  $buf = "";
  $mol = "";
  $txt = "";
  $lbl = "";
  $ct  = 1;
  if ($itemcount > 0) {
    $dbbasedir = $basedir . "/" . $dbnum;
    $dbbaseURLdir = $baseURLdir . "/" . $dbnum;
    if ( ! -d $dbbasedir) {
      mkdir $dbbasedir || die("ERROR: cannot create directory $dbbasedir !");
      if ($verbose > 1) { print "created subdirectory: $dbbasedir\n"; }
    }
    if ( ! -W $dbbasedir) {
      die("ERROR: cannot write to directory $dbbasedir !");
    }
    # first round: check for missing files, if $level > 2
    if ($level > 1) {
      if ($verbose > 0) {
        print "checking for missing files in data collection $dbnum\n";
      }
      $last_id = 0;
      for ($ii = 0; $ii < $nchunks; $ii++) {
        $qstr = "SELECT $idname, $struc3dfield FROM $datatable WHERE $idname > $last_id LIMIT 0, 100";
        $sth = $dbh->prepare($qstr);
        $sth->execute();
        $rows = $sth->rows();
        if ($rows > 0) {        
          while ($ref = $sth->fetchrow_hashref()) {
            $item_id   = $ref->{"$idname"};
            $last_id = $item_id;
            $struc3durl  = $ref->{"$struc3dfield"};
            if (length($struc3durl) > 0) {
              $fname = $item_id;
              while (length($fname) < $digits) { $fname = "0" . $fname; }
              $subdir = '';
              if ($subdirdigits > 0) {
                $subdir = substr($fname,0,$subdirdigits);
                $newdir = $dbbitmapdir . "/" . $subdir;
                $subdir = $subdir . "/";        
              }
              $fname = $dbbasedir . "/" . $subdir . $fname . ".mol";
              if ( ! -f $fname) { 
                if ($verbose > 1) { print " missing file: $fname\n"; }
                # now set the struc3d field in moldatatable to an empty string in order
                # to trigger (re-)creation of the 3D structure in the next step                
                $updstr = "UPDATE $datatable SET `$struc3dfield` = '' WHERE $idname = $item_id";
                $dbh->do($updstr);
              }
            }
          }     # while ...
        }       # if ($rows > 0)...
        $sth->finish;
      }         # for ...
    }           # if level > 2
    
    for ($ii = 0; $ii < $nchunks; $ii++) {
      # no offset for LIMIT as the searched item (status) is permanently updated
      $qstr = "SELECT ${structable}.$idname, ${structable}.struc FROM ";
      $qstr .= "$structable, $datatable WHERE ($struc3dfield = '') AND ";
      $qstr .= "(${structable}.$idname = ${datatable}.$idname) LIMIT 0, 100";
      $sth = $dbh->prepare($qstr);
      $sth->execute();
      while ($ref = $sth->fetchrow_hashref()) {
        $item_id   = $ref->{"$idname"};
        $struc     = $ref->{'struc'};
        $counter ++;
        $isvalid = 0;
        if ( $dbtype == 1) { $isvalid = valid_mol($struc); }
        if ($isvalid eq 1) {
          $fname = $item_id;
          while (length($fname) < $digits) { $fname = "0" . $fname; }
          $subdir = '';
          if ($subdirdigits > 0) {
            $subdir = substr($fname,0,$subdirdigits);
            $newdir = $dbbasedir . "/" . $subdir;
            if ( ! -d $newdir) {
              mkdir $newdir || die("ERROR: cannot create directory $newdir !");
              if ($verbose > 1) { print "created subdirectory: $newdir\n"; }
            }
            if ( ! -W $newdir) {
              die("ERROR: cannot write to directory $newdir !");
            }
            $subdir = $subdir . "/";        
          }
          $fullfname = $dbbasedir . "/" . $subdir . $fname . ".mol";
          $URLfname  = $dbbaseURLdir  . "/" . $subdir . $fname . ".mol";
          if ($verbose > 1) {
            print " creating $fullfname\n";
          }
          process_struc($struc,$fullfname,$options,$scale);

          if ( -f $fullfname) {
            # add this URL to the struc3dfield in moldatatable
            $updstr = "UPDATE $datatable SET `$struc3dfield` = \"$URLfname\" WHERE $idname = $item_id";
            $sth2 = $dbh->do($updstr);
          }
        } else {
          $counter --;
          $badmols ++;
        }
      }                 # end while ($ref...
      $sth->finish;
    }                   # end "for" loop
  }   # if itemcount > 0 ...    

  #===================================check for molfiles to be deleted
  if ($level > 2) {
    $dbbasedir = $basedir . "/" . $dbnum;
    $dbbaseURLdir = $baseURLdir . "/" . $dbnum;
    if ($verbose > 0) {
      print "checking for orphan mol files in $dbbasedir\n";
    }
    if ($ostype eq 1) { open(FIND,"find $dbbasedir -name \"*.mol\" -print |"); }
    if ($ostype eq 2) { 
      $dbbasedir_win = $dbbasedir;
  	$dbbasedir_win =~ s/\//\\/g;
  	$fullpath = $dbbasedir_win . "\\\*.mol";
  	open(FIND,"dir $fullpath /b /s |"); 
    }
    while ($fullname = <FIND>) {
      chomp($fullname);  
  	if ($ostype eq 2) { $fullname =~ s/\\/\//g; }
      $slashpos = rindex($fullname,"/");
      $fname = substr($fullname,($slashpos + 1));
      $item_id = $fname;
      $item_id =~ s/\.mol//g;
      # check if this id exists in our data collection
      $itemcount = 1;
      $sth0 = $dbh->prepare("SELECT COUNT($idname) AS itemcount FROM $datatable WHERE $idname = $item_id ");
      $sth0->execute();
      while ($ref0 = $sth0->fetchrow_hashref()) {
        $itemcount = $ref0->{'itemcount'};
      }
      $sth0->finish();
      if ($itemcount < 1) {
        if ($verbose > 1) {
          if ($reallydelete == 1) { 
            print " deleting file $fullname\n";
          } else {
            print " $fullname marked to be deleted\n";
          }
        }
        $delcounter++;
        if ($reallydelete == 1) { 
          (unlink $fullname) || print STDERR "ERROR: failed to delete $fullname\n";
        } else {
          $delname = $fullname . ".to_be_deleted";
          (rename $fullname, $delname) || print STDERR "ERROR: renaming failed for $fullname\n";
        }
      }
    }
  }
  #=================================end deletion

} # for ($idb ....


$dbh->disconnect();

if ($verbose > 0) {
  print "$counter records processed in total\n";
  print "$badmols records ignored\n";
  print "$delcounter mol file(s) deleted or marked to be deleted\n";
}


#===================== subroutines =======================================

sub valid_mol() {
  $testmol = shift;
  $zerolines = 0;
  @xyzline = split(/\n/, $testmol);
  for ($i = 0; $i <= $#xyzline; $i++) {
    $testline = $xyzline[$i];
    $testline =~ s/\ +/:/g;
    @xyz = split(/:/, $testline);
    $xval = $xyz[1];
    $yval = $xyz[2];
    $zval = $xyz[3];
    if ((index($xval,"0.0000") >= 0) && (index($yval,"0.0000") >= 0) && 
       (index($zval,"0.0000") >= 0)) { $zerolines ++; }
  }
  if ($zerolines > 1) {
    return 0;
  } else {
    return 1;
  }
}

sub valid_rxn() {
  my $testrxn = shift;
  my $result = 0;
  if ((index($testrxn,'$RXN') == 0) && (index($testrxn,'$MOL') > 0) && (index($testrxn,'M  END') > 0)) {
    $result = 1;
  }
  return($result);
}


sub process_struc() {
  $structure = shift;
  $filename = shift;
  $structure =~ s/\"/\\\"/g;
  $structure =~ s/\$/\\\$/g;
  if (index($structure,"M  END") < 0) { $structure = $structure . "M  END\n"; }	
  $tmpdir = File::Spec->tmpdir();
  ($tmp1fh, $tmp1filename) = tempfile(UNLINK => 1, DIR => $tmpdir, SUFFIX => ".mol");
  $input = $structure;
  $input =~ s/\\\$/\$/g;
  $input =~ s/\r//g;
  $input =~ s/\n/\r\n/g;
  print $tmp1fh "$input\n";
  $tmp2filename = $tmp1filename;
  $tmp2filename =~ s/\.mol/_UFF\.mol/g;
  $tmp3filename = $tmp1filename;
  $tmp3filename =~ s/\.mol/_MMFF94\.mol/g;
  #print "     $tmp1filename ... $tmp2filename\n";
  system("$OBMINIMIZE -h -sd -n 10000 -ff UFF -omdl $tmp1filename > $tmp2filename 2>/dev/null");
  system("$OBMINIMIZE -n 10000 -ff MMFF94 -omdl $tmp2filename > $tmp3filename 2>/dev/null");
  copy($tmp3filename,$filename) or print "Copy failed for $filename";
  close $tmp1fh;
  unlink $tmp2filename, $tmp3filename;
}


sub filterthroughcmd {
  $input   = shift;
  $cmd     = shift;
  open(FHSUB, "echo \"$input\"|$cmd 2>&1 |");   # stderr must be redirected to stdout
  $res      = "";                               # because the Ghostscript "bbox" device
  while($line = <FHSUB>) {                      # writes to stderr
    $res = $res . $line;
  }
  return $res;
}

sub filterthroughcmd2 {                         # workaround for Windows 
  $input   = shift;
  $cmd     = shift;
  ($tmpfh, $tmpfilename) = tempfile(UNLINK => 1);
  $input =~ s/\\\$/\$/g;
  $input =~ s/\r//g;
  $input =~ s/\n/\r\n/g;
  print $tmpfh "$input\n";
  #open(FHSUB, "type $tmpfilename |$cmd 2>&1 |");   # stderr must be redirected to stdout
  open(FHSUB, "$cmd < $tmpfilename 2>&1 |");   # stderr must be redirected to stdout
  $res      = "";                               # because the Ghostscript "bbox" device
  while($line = <FHSUB>) {                      # writes to stderr
    $res = $res . $line;
  }
  close $tmpfh;
  return $res;
}

sub ltrim() {
  $subline1 = shift;
  $subline1 =~ s/^\ +//g;
  return $subline1;
}

sub getostype() {
  $os  = "";
  $osresult = 1;
  $os  = uc($ENV{OS});
  if ($os eq "") { $os = uc($ENV{OSTYPE}); }
  if (index($os,"WINDOWS")>=0) {
    $osresult = 2;
  }
  return $osresult;
}

sub getval() {
  my $argline = shift;
  chomp($argline);
  my @argpart = split(/=/, $argline);
  my $argval = "";
  if ($#argpart > 0) { $argval = $argpart[1]; }
  return $argval;
}
