#!/usr/bin/perl
#
# mkautomol.pl  Norbert Haider, University of Vienna, 2014-2018
#               norbert.haider@univie.ac.at
#
# This script is part of the MolDB6 package. Last change: 2018-01-05
#
# Example script which (re-)generates the auto_mol_formula,
# auto_mol_fw (and optionally the auto_mol_inchikey) columns in the
# moldata MySQL table within the MolDB6 database.

$level    = 1;  # 0 = dry-run, 1 = fix only missing entries, 2 = fix all entries
$verbose  = 1;  # 0 = silent operation, 
                # 1 = report each data collection, 
                # 2 = report each molecule

# some default settings that will be overridden by the configuration file
$enable_inchi = "y";
$INCHI        = "/usr/local/bin/inchi-1";
$INCHI_OPT    = "-STDIO -Key";

use DBI();
$ostype = getostype();
if ($ostype eq 2) { use File::Temp qw/ tempfile tempdir /; }

$configfile = "../moldb6.conf";

$return     = do $configfile;
if (!defined $return) {
  die ("ERROR: cannot read configuration file $configfile!\n");
}	

$user     = $rw_user;    # from configuration file
$password = $rw_password;

$dbh = DBI->connect("DBI:mysql:database=$database;host=$hostname",
                    $user, $password,
                    {'RaiseError' => 1});
$qstr = "SET SESSION sql_mode = '';";
$dbh->do($qstr);

# read moldb_meta table and find out which data collections are to be processed

$ndb = 0;
@db = [""];
$sth0 = $dbh->prepare("SELECT db_id FROM $metatable WHERE type = 1 ORDER BY db_id");
$sth0->execute();
while ($ref0 = $sth0->fetchrow_hashref()) {
  $db_id = $ref0->{'db_id'};
  $ndb++;
  @db[($ndb-1)] = $db_id;
}
$sth0->finish;

$nfixed = 0;  # counter for all entries that were fixed

for ($i = 0; $i < $ndb; $i++) {
  $dbnum = @db[$i];
  $dbprefix = $prefix . "db" . $dbnum . "_";
  $moldatatable  = $dbprefix . $moldatasuffix;
  $molstructable = $dbprefix . $molstrucsuffix;
  
  if ($verbose > 0) {
    print "processing data collection $dbnum: (re-)creating auto_mol_ fields in table $moldatatable ";
  }
  
  # examine the structure of moldatatable to find out which fields are to be filled
  $qstr = "SHOW FULL COLUMNS FROM $moldatatable";
  $sth1 = $dbh->prepare($qstr);
  $sth1->execute();
  $found_auto_mol_formula = 0;
  $found_auto_mol_fw = 0;
  $found_auto_mol_inchikey = 0;
  while ($ref1 = $sth1->fetchrow_hashref()) {
    $field =  $ref1->{'Field'};
    $type =  $ref1->{'Type'};
    if ($field eq "auto_mol_formula") { $found_auto_mol_formula = 1; }
    if ($field eq "auto_mol_fw") { $found_auto_mol_fw = 1; }
    if ($field eq "auto_mol_inchikey") { $found_auto_mol_inchikey = 1; }
  }
  $sth1->finish;

  # read all molecules from moldatatable/molstructable and pipe the MDL molfiles
  # through checkmol (and inchi-1) to get the formula and FW (and InChiKey)
  #
  # first, get the number of rows and chop the whole operation into
  # suitable chunks
  $sth0 = $dbh->prepare("SELECT COUNT(mol_id) AS molcount FROM $moldatatable ");
  $sth0->execute();
  while ($ref0 = $sth0->fetchrow_hashref()) {
    $molcount = $ref0->{'molcount'};
  }
  $sth0->finish();
  if ($verbose > 0) {
    print " ($molcount molecules)\n";
  }
  $nchunks = int( (($molcount + 999) / 1000) );

  $whatstr = "";
  if ($found_auto_mol_formula gt 0) {
    $whatstr = "`auto_mol_formula`"; 
  }  
  if ($found_auto_mol_fw > 0) {
    if (length($whatstr) > 0) { $whatstr .= ", "; }
    $whatstr .= "`auto_mol_fw`"; 
  }  
  if ($found_auto_mol_inchikey > 0) {
    if (length($whatstr) > 0) { $whatstr .= ", "; }
    $whatstr .= "`auto_mol_inchikey`"; 
  } 
  if (length($whatstr) > 0) {  
    $last_id = 0;
    for ($j = 0; $j < $nchunks; $j++) {
      #$offset = $j * 1000;
      #$sth = $dbh->prepare("SELECT `mol_id`, $whatstr FROM $moldatatable LIMIT $offset,1000");
      $sth = $dbh->prepare("SELECT `mol_id`, $whatstr FROM $moldatatable WHERE mol_id > $last_id LIMIT 0,1000");
      $sth->execute();
      while ($ref = $sth->fetchrow_hashref()) {
        $missing_formula = 0;
        $missing_fw = 0;
        $missing_inchikey = 0;
        $mol_id = $ref->{'mol_id'};
        $last_id = $mol_id;
        if ($found_auto_mol_formula > 0) {
          $formula = $ref->{'auto_mol_formula'};
          if ((length($formula) == 0) || ($level == 2)) { $missing_formula = 1; }
        }
        if ($found_auto_mol_fw > 0) {
          $fw = $ref->{'auto_mol_fw'};
          if ((length($fw) == 0) || ($fw < 1) || ($level == 2)) { $missing_fw = 1; }
        }
        if ($found_auto_mol_inchikey > 0) {
          $inchikey = $ref->{'auto_mol_inchikey'};
          if (((length($inchikey) == 0) || ($level == 2)) && ($enable_inchi eq "y")) { $missing_inchikey = 1; }
        }

        if (($missing_formula > 0) || ($missing_fw > 9) || ($missing_inchikey > 0)) {
          if ($verbose > 1) {
            print "  mol_id $mol_id in table $moldatatable needs to be fixed: ";
            if ($missing_formula > 0) { print "auto_mol_formula "; }       
            if ($missing_fw > 0) { print "auto_mol_fw "; }       
            if ($missing_inchikey > 0) { print "auto_mol_inchikey "; }       
            print "\n";
          }

          # now get the molfile and derive the missing values
          $sth1 = $dbh->prepare("SELECT struc FROM $molstructable WHERE mol_id = $mol_id");
          $sth1->execute();
          $molfile = "";
          while ($ref1 = $sth1->fetchrow_hashref()) {
            $molfile = $ref1->{'struc'};
          }
          $sth1->finish();
          $updstr = "";
          if (($missing_formula > 0) || ($missing_fw > 0)) {
            my $mfmw = get_mfmw($molfile);
            # print "$mfmw\n";
            # if unsuccessful, pipe the molfile again through checkmol
            if ($mfmw eq ":") {
              if ($ostype eq 2) {
                $molfile = filterthroughcmd2($molfile,"$CHECKMOL -m -");
              } else {
                $molfile = filterthroughcmd($molfile,"$CHECKMOL -m -");
              }
              my $mfmw = get_mfmw($molfile);
            }
            my @amfmw = split (/:/, $mfmw);
            my $mf = $amfmw[0];
            chomp($mf);
            $mf = rtrim($mf);
            my $mw = $amfmw[1];
            chomp($mw);
            if ($missing_formula > 0) {
              $updstr = "`auto_mol_formula` = '$mf'";
              $nfixed++;
            }
            if ($missing_fw > 0) {
              if (length($updstr) > 0) { $updstr .= ", "; }
              $updstr .= "`auto_mol_fw` = '$mw'";
              $nfixed++;
            }
          }
          if ($missing_inchikey > 0) {
            if ($ostype eq 2) {
              $inchi  = filterthroughcmd2($molfile,"$INCHI $INCHI_OPT");
            } else {
              $inchi  = filterthroughcmd($molfile,"$INCHI $INCHI_OPT");
            }
            @inchi_arr = split(/InChIKey=/,$inchi);
            $inchi = $inchi_arr[1];
            @inchi_arr = split(/\n/,$inchi);
            $inchikey = $inchi_arr[0];
            chomp($inchikey);
            if (length($updstr) > 0) { $updstr .= ", "; }
            $updstr .= "`auto_mol_inchikey` = '$inchikey'";
            $nfixed++;
          }
          if (length($updstr) > 0) {
            $updstr = "UPDATE $moldatatable SET " . $updstr . " WHERE mol_id = $mol_id";
            if ($level > 0) {
              $dbh->do($updstr);
            }
          }
        }

      }                                    # end while ($ref...
      $sth->finish;
    }  # end "for ($j" loop
  }    # if length($whatstr)


}   # end "for ($i.." loop

$dbh->disconnect();

if ($verbose > 0) {
  if ($level == 0) {
    print "  values to be fixed: $nfixed\n";
  } else {
    print "  values fixed: $nfixed\n";
  }
}

#============================================================

sub filterthroughcmd {
  my $input   = shift;
  my $cmd     = shift;
  open(FHSUB, "echo \"$input\"|$cmd 2>&1 |");
  $output     = '';
  $res        = "";
  while($line = <FHSUB>) {
    $res = $res . $line;
  }
  return $res;
}

sub filterthroughcmd2 {                         # workaround for Windows 
  $input   = shift;
  $cmd     = shift;
  ($tmpfh, $tmpfilename) = tempfile(UNLINK => 1);
  $input =~ s/\\\$/\$/g;
  $input =~ s/\r//g;
  $input =~ s/\n/\r\n/g;
  print $tmpfh "$input\n";
  #open(FHSUB, "type $tmpfilename |$cmd 2>&1 |");   # stderr must be redirected to stdout
  open(FHSUB, "$cmd < $tmpfilename 2>&1 |");   # stderr must be redirected to stdout
  $res      = "";                               # because the Ghostscript "bbox" device
  while($line = <FHSUB>) {                      # writes to stderr
    $res = $res . $line;
  }
  close $tmpfh;
  return $res;
}

sub getostype() {
  $os  = "";
  $osresult = 1;
  $os  = uc($ENV{OS});
  if ($os eq "") { $os = uc($ENV{OSTYPE}); }
  if (index($os,"WINDOWS")>=0) {
    $osresult = 2;
  }
  return $osresult;
}

sub get_mfmw() {
  my $inmol = shift;
  my @marr = split (/\n/, $inmol );
  my $mfmwline = $marr[2];
  my @mfmwa1 = split (/MW=/,$mfmwline);
  my $mw = $mfmwa1[1];
  chomp($mw);
  my $mf1 = $mfmwa1[0];
  my @mfmwa2 = split (/MF=/,$mf1);
  my $mf = $mfmwa2[1];
  $retval = "$mf" . ":" . "$mw";
  return $retval;
}

sub ltrim() {
  $subline1 = shift;
  $subline1 =~ s/^\ +//g;
  return $subline1;
}

sub rtrim() {
  $subline2 = shift;
  $subline2 =~ s/\ +$//g;
  return $subline2;
}
