#!/usr/bin/perl
#
# sdf2sql.pl     Norbert Haider, University of Vienna, 2015-2021
#                norbert.haider@univie.ac.at
#
# This script is part of the MolDB6 package. Last change: 2021-02-28
#
# This script reads an SDF file which was previously analyzed by the script 
# "sdfcheck.pl" and prepares an intermediate SQL file that can be later
# added into a MySQL-based MolDB6 data collection. The auxiliary tables are 
# also created and SVG graphics are created. Bitmap graphics are NOT created.
#
# ===========================================================================
#
# Purpose: setting up a web-based, fully searchable molecular structure 
# database (MolDB6); for a more detailed description see
# https://homepage.univie.ac.at/norbert.haider/cheminf/moldb.html
#
# This software was developed and tested with the SDF files from the 
# PubChem database, which are freely available at
# ftp://ftp.ncbi.nlm.nih.gov/pubchem/Compound/CURRENT-Full/SDF/
#
# Requirement for making SVG graphics: "mol2svg" (as part of mol2ps/mol2svg,
# see below) version 0.4 or higher must be installed.
#

$verbose          = 0;   # 0, 1 or 2
$use_fixed_fields = 0;   # 0 or 1, should be 1 for older versions of checkmol
$chkzero          = 1;   # 0 or 1, check if molecule contains > 1 atoms with X,Y,Z = 0,0,0
$create_table     = 1;   # 0 or 1, include "CREATE TABLE ... IF NOT EXISTS" statements
$fpfile           = "../fp01.sdf";
$configfile       = "../moldb6.conf";    # needs only a subset of settings (no passwords etc.)
$deffile          = "sdf2moldb.def";
$keep_comments    = 0;   # 0 = delete comment lines; 1 = leave these lines unchanged


# define some defaults, may be overridden by config file

$fpdict_mode      = 1;  # 1 = auto adjust, 2 = force 64 bit, 3 = force 32 bit
$MOL2SVG         = "/usr/local/bin/mol2svg";
$mol2svgopt      = "--rotate=auto3Donly --hydrogenonmethyl=off --color=/usr/local/etc/color.conf"; # options for mol2svg, e.g. "--showmolname=on"
#$mol2svgopt     = "--rotate=auto3Donly --hydrogenonmethyl=off"; # options for mol2svg, e.g. "--showmolname=on"
$svg_scalingfactor = 1.0;           # 1.0  gives good results
$tweakmolfiles   = "y";
$enable_inchi    = "y"; # "y" or "n"; creates "mol_auto_inchikey" column in every new moldata table
$INCHI           = "/usr/local/bin/inchi-1";
$INCHI_OPT       = "-STDIO -Key"; # Linux: " -STDIO -Key", Windows: "/STDIO /Key /AuxNone"


if ($#ARGV < 0) {
  print "Usage: sdf2sql.pl <inputfile>\n";
  exit;
}

$infile = $ARGV[0];

$ostype = getostype();
if ($ostype eq 2) { use File::Temp qw/ tempfile tempdir /; }

$return     = do $configfile;
if (!defined $return) {
  die("ERROR: cannot read configuration file $configfile!\n");
}	

# use the following settings to override the values from the configuration file

# assuming mol2svg version 0.4 or higher: scaling is added to the options, 
# if not already there
if (index($mol2svgopt, '--scaling=') < 0) {
  $mol2svgopt .= " --scaling=" . $svg_scalingfactor; 
}


$mysql_charset   = "latin1";          # default
$mysql_collation = "latin1_swedish_ci";

if ($charset eq "latin2") {           # from configuration file
  $mysql_charset   = "latin2";
  $mysql_collation = "latin2_general_ci";
}
if ($charset eq "utf8") {
  $mysql_charset   = "utf8";
  $mysql_collation = "utf8_unicode_ci";
}

$db_id                   = 1;  # default data collection number
$db_access               = 1;  # default access mode (0 = hidden, 1 = read-only, 2 = read/write)
$bitmapfile_digits       = 8;  # default values for bitmap filenames (e.g., 00000123.png)
$bitmapfile_subdirdigits = 4;  # default value for subdirectory names (e.g., 0000/00000123.png)
$db_name                 = "";
$flags                   = 0;  # 32bit integer (unsigned); currently unused

$found_auto_mol_formula  = 1;  # default: 1
$found_auto_mol_fw       = 1;  # default: 1   
$found_auto_mol_inchikey = 0;  # default: 0

$found_mol_id   = 0;
$found_mol_name = 0;

open (DEF, "<$deffile") || die("ERROR: cannot open definition file $deffile!");
$nfields = 0;
while ($line = <DEF>) {
  chomp($line);
  @valid = split (/#/, $line);  # ignore everything behind the first pound sign
  $line  = $valid[0];
  $line  = ltrim($line);
  $line  = rtrim($line);
  if ((index($line,'sdfilename') == 0) && (index($line,'=') >= 10)) {
    @defrec = split (/=/, $line);
    $sdfile = $defrec[1];
    $sdfile = ltrim($sdfile);
  }
  if ((index($line,'db_id') == 0) && (index($line,'=') >= 5)) {
    @defrec = split (/=/, $line);
    $db_id = $defrec[1];
    $db_id = ltrim($db_id);
  }
  if ((index($line,'db_name') == 0) && (index($line,'=') >= 7)) {
    @defrec = split (/=/, $line);
    $db_name = $defrec[1];
    $db_name = ltrim($db_name);
    $db_name =~ s/\"//g;
  }
  if ((index($line,'db_description') == 0) && (index($line,'=') >= 14)) {
    @defrec = split (/=/, $line);
    $db_description = $defrec[1];
    $db_description = ltrim($db_description);
    $db_description =~ s/\"//g;
  }
  if ((index($line,'db_access') == 0) && (index($line,'=') >= 9)) {
    @defrec = split (/=/, $line);
    $tmp_access = $defrec[1];
    $tmp_access = ltrim($tmp_access);
    if ($tmp_access == 0) { $db_access = 0; }
    if ($tmp_access == 1) { $db_access = 1; }
    if ($tmp_access == 2) { $db_access = 2; }
    if ($tmp_access == 3) { $db_access = 3; }    
  }
  if ((index($line,'bitmapfile_digits') == 0) && (index($line,'=') >= 17)) {
    @defrec = split (/=/, $line);
    $tmp_digits = $defrec[1];
    $tmp_digits = ltrim($tmp_digits);
    if (($tmp_digits > 0) && ($tmp_digits <= 32)) { $bitmapfile_digits = $tmp_digits; }
  }
  if ((index($line,'bitmapfile_subdirdigits') == 0) && (index($line,'=') >= 23)) {
    @defrec = split (/=/, $line);
    $tmp_subdirdigits = $defrec[1];
    $tmp_subdirdigits = ltrim($tmp_subdirdigits);
    if (($tmp_subdirdigits >= 0) && ($tmp_subdirdigits <= $bitmapfile_digits)) { $bitmapfile_subdirdigits = $tmp_subdirdigits; }
  }
  if ((index($line,'flags') == 0) && (index($line,'=') >= 4)) {
    @defrec = split (/=/, $line);
    $tmp_flags = $defrec[1];
    $tmp_flags = ltrim($tmp_flags);
    chomp($tmp_flags);
    if (($tmp_flags + 0 eq $tmp_flags ) && ( $tmp_flags >= 0 ) && ( $tmp_flags < 2147483648 ) ) {
      $flags = $tmp_flags;
    } else {
      $flags = 0;
    }
  }
  
  $lpos = index($line,':');
  $rpos = rindex($line,':');
  if (($lpos >= 1) && ($rpos >= 3) && ($rpos > $lpos)) {
    # this should be a definition line
    @defrec = split (/:/, $line);
    $sdf_label   = $defrec[0];
    $mysql_label = $defrec[1];
    $mysql_type  = $defrec[2];
    $html_label  = $defrec[3];
    $html_format = $defrec[4];
    if (index($mysql_label,"auto_mol_") != 0) {
      @afield[($nfields)] = [ $sdf_label, $mysql_label, $mysql_type, $html_label, $html_format, "" ];
      $nfields++;
    } else {
      if (index($mysql_label,"auto_mol_formula") == 0) { $found_auto_mol_formula = 1; }
      if (index($mysql_label,"auto_mol_fw") == 0) { $found_auto_mol_fw = 1; }  
      if (index($mysql_label,"auto_mol_inchikey") == 0) { 
        if ($enable_inchi == "y") { $found_auto_mol_inchikey = 1; }
      }  

    }
    if ($mysql_label eq "mol_name") { $found_mol_name = 1; }
    if ($mysql_label eq "mol_id") { $found_mol_id = 1; print "$line\n"; }
  }
}

if ($verbose > 0) { 
  print "\n-- Adding data to MolDB6 database '$database', using configuration\n";
  print "-- file '$configfile'.\n\n";
  print "-- Your SDF input file must have the same format as the one which\n";
  print "-- has been used previously for analysis with the sdfcheck.pl script:\n";
  print "-- $sdfile\n\n";
}

if ($found_mol_id > 0) {
  print STDERR "\nERROR: your definition file $deffile contains a field named\n";
  print STDERR "'mol_id'! This label must not be used, as it is automatically created\n";
  print STDERR "by the system. Please edit $deffile and change 'mol_id' into some\n";
  print STDERR "other name.\n";
  exit;
}

if ($found_mol_name == 0) {
  print "\n-- WARNING: your definition file $deffile does not contain a field\n";
  print "-- named 'mol_name'. It is highly recommended to rename the most descriptive\n";
  print "-- field into 'mol_name', please read the instructions in $deffile!\n";
  print "-- Otherwise, a field 'mol_name' will be automatically created, but it\n";
  print "-- might remain empty during the SDF import operation.\n\n";
}


if ($verbose > 1) { 
  print "-- using SD file $infile\n"; 
  print "-- found the following data field definitions:\n";
  for ($i = 0; $i <= $#afield; $i++) {
    $l1  = $afield[$i][0];
    $l2  = $afield[$i][1];
    $l3  = $afield[$i][2];
    $l4  = $afield[$i][3];
    $l5  = $afield[$i][4];
    while (length($l1) < 25) { $l1 = $l1 . " "; }
    while (length($l2) < 25) { $l2 = $l2 . " "; }
    while (length($l3) < 25) { $l3 = $l3 . " "; }
    while (length($l4) < 5) { $l4 = $l4 . " "; }
    print "-- $l1   $l2  $l3 $l4 $l5\n";
  } 
}

if ($verbose > 0) { 
  print "\n-- these settings are used:\n";
  print "--  MySQL database:           $database\n";
  print "--  SDF file:                 $infile\n";
  print "--  data collection number:   $db_id\n";
  print "--  data collection name:     $db_name\n";
  print "--  description:              $db_description\n";
  print "--  access mode (0,1,2,3):    $db_access\n";
  print "--  bitmap directory:         $bitmapdir\n";
  print "--  bitmap filename digits:   $bitmapfile_digits\n";
  print "--  subdirectory name digits: $bitmapfile_subdirdigits\n\n";
  print "--  flags:                    $flags\n\n";  
}

# check for checkmol
$return = `$CHECKMOL -v`;
if (index($return,"Usage:") < 0) {
  die("ERROR: could not find 'checkmol', make sure it is installed");	
}  

# check for matchmol
$return = `$MATCHMOL -v`;
if (index($return,"Usage:") < 0) {
  die("ERROR: could not find 'matchmol', make sure it is installed");	
}  

# check for mol2svg
$return = `$MOL2SVG -v`;
if (index($return,"Usage:") < 0) {
  die("ERROR: could not find 'mol2svg', make sure it is installed");	
}  


open (FPF, "<$fpfile") || die ("ERROR: cannot open fingerprint file $fpfile!");

$fpbuf = "";
$n_dict  = 0;
$n_mol   = 0;
$molbuf  = "";
$fptype  = 1;  # 1 = 64-bit, 2 = 32-bit
if ($fpdict_mode < 3) { $max_bs = 62; } else { $max_bs = 31; }
@fpstruc = [""];
@fptypes = [""];

while ($line = <FPF>) {
  if ((substr($line,0,4) eq '$$$$') || eof(FPF)) {
    if (length($molbuf)>20) {  
      if (length($fpbuf) > 20) { $fpbuf .= '$$$$ ' . "\r\n"; }
      if (index($molbuf,"M  END")<0) { $molbuf .= "M  END\r\n"; }
      $fpbuf .= $molbuf;
      $molbuf = "";
      $n_mol++;
    }
    if (($n_mol % $max_bs == 0) || eof(FPF)) {
      # insert dictionary chunk into fpdef table
      if ($n_mol > 31) {
      	$fptype = 1;
      } else {
        if ($fpdict_mode == 2) { $fptype = 1; } else { $fptype = 2; }
      }
      $n_dict++;
      #print "$fpbuf\n";
      #print "chunk $n_dict contains $n_mol structures (type $fptype)\n";
      @fpstruc[($n_dict - 1)] = $fpbuf;
      @fptypes[($n_dict - 1)] = $fptype;
      $fpbuf = "";
      $molbuf = "";
      $n_mol = 0;
    }
  } else {
    $molbuf = $molbuf . $line;
  }
}
close(FPF);

$dbprefix      = $prefix . "db" . $db_id . "_";
$molstructable = $dbprefix . $molstrucsuffix;
$moldatatable  = $dbprefix . $moldatasuffix;
$molstattable  = $dbprefix . $molstatsuffix;
$molfgbtable   = $dbprefix . $molfgbsuffix;
$molcfptable   = $dbprefix . $molcfpsuffix;
$pic2dtable    = $dbprefix . $pic2dsuffix;

open (SDF, "<$infile") || die("ERROR: cannot open input file $infile!");

# read the fragment dictionary from the fpstruc array

$createstr = "";
for ($f = 0; $f < $n_dict; $f++ ) {
  $fpdef = @fpstruc[$f];
  $fptype = @fptypes[$f];
  if (length($fpdef)>20) {
    $dictnum = $f + 1;
    while (length($dictnum) < 2) { $dictnum = "0" . $dictnum;  }
    if ($fptype == 1) {
      $createstr .= "  dfp$dictnum BIGINT NOT NULL,\n";
    } else {
      $createstr .= "  dfp$dictnum INT(11) UNSIGNED NOT NULL,\n";
    }
  }
}                 # end for

chomp($createstr);
if ($n_dict < 1) {
  die("ERROR: could not retrieve fingerprint definition from the dictionary file\n");
}

# create tables if they do not exist already

$createcmd = "CREATE TABLE IF NOT EXISTS $molstructable (
  mol_id INT(11) NOT NULL DEFAULT '0', 
  struc MEDIUMBLOB NOT NULL,
  PRIMARY KEY mol_id (mol_id)
) ENGINE=MYISAM CHARACTER SET $mysql_charset COLLATE $mysql_collation COMMENT='Molecular structures';";
if ($create_table == 1) {
  print "$createcmd\n";
} else {
  print "-- XXXXc-\n";  # insert a code to signal that CREATE TABLE statements are missing
}

$createcmd = "CREATE TABLE IF NOT EXISTS $moldatatable (
  mol_id INT(11) NOT NULL DEFAULT '0',\n" ;
if ($found_mol_name == 0) {
  $createcmd = $createcmd . "  mol_name TEXT NOT NULL,\n";
}  
# create auto_mol_ fields
$createcmd = $createcmd . "  auto_mol_formula VARCHAR(250) NOT NULL COMMENT '>>>>Formula<3<auto_mol_formula<0<<',\n";
$createcmd = $createcmd . "  auto_mol_fw DOUBLE NOT NULL DEFAULT '0.0' COMMENT '>>>>MW<1<auto_mol_fw<0<<',\n";
if ($enable_inchi == "y") {
  $createcmd = $createcmd . "  auto_mol_inchikey VARCHAR(250) NOT NULL COMMENT '>>>>InChIKey<1<inchikey<1<<',\n";
}

for ($i = 0; $i <= $#afield; $i++) {
  $l1  = $afield[$i][0];       # SDF field name
  $l2  = $afield[$i][1];       # MySQL field name
  $l3  = $afield[$i][2];       # MySQL field type
  $l4  = $afield[$i][3];       # HTML field name
  $l5  = $afield[$i][4];       # HTML format option
  $l6  = $afield[$i][5];       # search mode
  $l7  = $afield[$i][6];       # reserved
  if ($l4 eq "") { $l4 = $l2; }
  if ($l5 eq "") { $l5 = 1; }
  if (index($l1,"auto_mol_") != 0) {
    $createcmd = $createcmd . "  $l2 $l3 COMMENT '>>>>$l4<$l5<$l1<$l6<$l7<',\n";
  }
} 
$createcmd = $createcmd . "  PRIMARY KEY mol_id (mol_id)
) ENGINE=MYISAM CHARACTER SET $mysql_charset COLLATE $mysql_collation COMMENT='Molecular data'";
if ($create_table == 1) {
  print "$createcmd ;\n";
}

$createcmd = "CREATE TABLE IF NOT EXISTS $molstattable (
  mol_id int(11) NOT NULL DEFAULT '0', \n";
open (MSDEF, "$CHECKMOL -l |");
$nfields = 0;
while ($line = <MSDEF>) {
  chomp($line);
  @valid = split (/:/, $line);  # ignore everything behind the colon
  $line  = $valid[0];
  if (index($line,'n_') == 0) {
    $nfields ++;
    $createcmd = $createcmd . "  $line" . " SMALLINT(6) NOT NULL DEFAULT '0',\n";
  }
}
$createcmd = $createcmd . "  PRIMARY KEY  (mol_id)
) ENGINE=MYISAM COMMENT='Molecular statistics';";

# check if we have an older version of checkmol; if yes,
# revert to fixed field definitions
if (($nfields < 50) || ($use_fixed_fields == 1)) {
  if ($verbose > 0) {
    print "-- reverting to fixed molstat fields\n";
  }
  $createcmd="CREATE TABLE IF NOT EXISTS $molstattable (
  mol_id int(11) NOT NULL DEFAULT '0',
  n_atoms SMALLINT(6) NOT NULL DEFAULT '0',
  n_bonds SMALLINT(6) NOT NULL DEFAULT '0',
  n_rings SMALLINT(6) NOT NULL DEFAULT '0',
  n_QA SMALLINT(6) NOT NULL DEFAULT '0',
  n_QB SMALLINT(6) NOT NULL DEFAULT '0',
  n_chg SMALLINT(6) NOT NULL DEFAULT '0',
  n_C1 SMALLINT(6) NOT NULL DEFAULT '0',
  n_C2 SMALLINT(6) NOT NULL DEFAULT '0',
  n_C SMALLINT(6) NOT NULL DEFAULT '0',
  n_CHB1p SMALLINT(6) NOT NULL DEFAULT '0',
  n_CHB2p SMALLINT(6) NOT NULL DEFAULT '0',
  n_CHB3p SMALLINT(6) NOT NULL DEFAULT '0',
  n_CHB4 SMALLINT(6) NOT NULL DEFAULT '0',
  n_O2 SMALLINT(6) NOT NULL DEFAULT '0',
  n_O3 SMALLINT(6) NOT NULL DEFAULT '0',
  n_N1 SMALLINT(6) NOT NULL DEFAULT '0',
  n_N2 SMALLINT(6) NOT NULL DEFAULT '0',
  n_N3 SMALLINT(6) NOT NULL DEFAULT '0',
  n_S SMALLINT(6) NOT NULL DEFAULT '0',
  n_SeTe SMALLINT(6) NOT NULL DEFAULT '0',
  n_F SMALLINT(6) NOT NULL DEFAULT '0',
  n_Cl SMALLINT(6) NOT NULL DEFAULT '0',
  n_Br SMALLINT(6) NOT NULL DEFAULT '0',
  n_I SMALLINT(6) NOT NULL DEFAULT '0',
  n_P SMALLINT(6) NOT NULL DEFAULT '0',
  n_B SMALLINT(6) NOT NULL DEFAULT '0',
  n_Met SMALLINT(6) NOT NULL DEFAULT '0',
  n_X SMALLINT(6) NOT NULL DEFAULT '0',
  n_b1 SMALLINT(6) NOT NULL DEFAULT '0',
  n_b2 SMALLINT(6) NOT NULL DEFAULT '0',
  n_b3 SMALLINT(6) NOT NULL DEFAULT '0',
  n_bar SMALLINT(6) NOT NULL DEFAULT '0',
  n_C1O SMALLINT(6) NOT NULL DEFAULT '0',
  n_C2O SMALLINT(6) NOT NULL DEFAULT '0',
  n_CN SMALLINT(6) NOT NULL DEFAULT '0',
  n_XY SMALLINT(6) NOT NULL DEFAULT '0',
  n_r3 SMALLINT(6) NOT NULL DEFAULT '0',
  n_r4 SMALLINT(6) NOT NULL DEFAULT '0',
  n_r5 SMALLINT(6) NOT NULL DEFAULT '0',
  n_r6 SMALLINT(6) NOT NULL DEFAULT '0',
  n_r7 SMALLINT(6) NOT NULL DEFAULT '0',
  n_r8 SMALLINT(6) NOT NULL DEFAULT '0',
  n_r9 SMALLINT(6) NOT NULL DEFAULT '0',
  n_r10 SMALLINT(6) NOT NULL DEFAULT '0',
  n_r11 SMALLINT(6) NOT NULL DEFAULT '0',
  n_r12 SMALLINT(6) NOT NULL DEFAULT '0',
  n_r13p SMALLINT(6) NOT NULL DEFAULT '0',
  n_rN SMALLINT(6) NOT NULL DEFAULT '0',
  n_rN1 SMALLINT(6) NOT NULL DEFAULT '0',
  n_rN2 SMALLINT(6) NOT NULL DEFAULT '0',
  n_rN3p SMALLINT(6) NOT NULL DEFAULT '0',
  n_rO SMALLINT(6) NOT NULL DEFAULT '0',
  n_rO1 SMALLINT(6) NOT NULL DEFAULT '0',
  n_rO2p SMALLINT(6) NOT NULL DEFAULT '0',
  n_rS SMALLINT(6) NOT NULL DEFAULT '0',
  n_rX SMALLINT(6) NOT NULL DEFAULT '0',
  n_rar SMALLINT(6) NOT NULL DEFAULT '0',
  PRIMARY KEY  (mol_id)
  ) ENGINE=MYISAM COMMENT='Molecular statistics';";
}
if ($create_table == 1) {
  print "$createcmd\n";
}

# create a new molfgb table
$createcmd="CREATE TABLE IF NOT EXISTS $molfgbtable (mol_id INT(11) NOT NULL DEFAULT '0', 
fg01 INT(11) UNSIGNED NOT NULL,
fg02 INT(11) UNSIGNED NOT NULL,
fg03 INT(11) UNSIGNED NOT NULL,
fg04 INT(11) UNSIGNED NOT NULL,
fg05 INT(11) UNSIGNED NOT NULL,
fg06 INT(11) UNSIGNED NOT NULL,
fg07 INT(11) UNSIGNED NOT NULL,
fg08 INT(11) UNSIGNED NOT NULL,
n_1bits SMALLINT NOT NULL,
PRIMARY KEY mol_id (mol_id)) ENGINE=MYISAM COMMENT='Functional group patterns';";
if ($create_table == 1) {
  print "$createcmd\n";
}
  
# create a new molcfp table if not yet present
$createcmd="CREATE TABLE IF NOT EXISTS $molcfptable (
  mol_id INT(11) NOT NULL DEFAULT '0',
$createstr
  hfp01 INT(11) UNSIGNED NOT NULL,
  hfp02 INT(11) UNSIGNED NOT NULL,
  hfp03 INT(11) UNSIGNED NOT NULL,
  hfp04 INT(11) UNSIGNED NOT NULL,
  hfp05 INT(11) UNSIGNED NOT NULL,
  hfp06 INT(11) UNSIGNED NOT NULL,
  hfp07 INT(11) UNSIGNED NOT NULL,
  hfp08 INT(11) UNSIGNED NOT NULL,
  hfp09 INT(11) UNSIGNED NOT NULL,
  hfp10 INT(11) UNSIGNED NOT NULL,
  hfp11 INT(11) UNSIGNED NOT NULL,
  hfp12 INT(11) UNSIGNED NOT NULL,
  hfp13 INT(11) UNSIGNED NOT NULL,
  hfp14 INT(11) UNSIGNED NOT NULL,
  hfp15 INT(11) UNSIGNED NOT NULL,
  hfp16 INT(11) UNSIGNED NOT NULL,
  n_h1bits SMALLINT NOT NULL, PRIMARY KEY (mol_id) 
  ) ENGINE=MYISAM COMMENT='Combined dictionary-based and hash-based fingerprints';";
if ($create_table == 1) {
  print "$createcmd\n";
}

# create a new pic2d table
$createcmd="CREATE TABLE IF NOT EXISTS $pic2dtable (
mol_id INT(11) NOT NULL DEFAULT '0',
type TINYINT NOT NULL DEFAULT '1' COMMENT '1 = png',
status TINYINT NOT NULL DEFAULT '0' COMMENT '0 = does not exist, 1 = OK, 2 = OK, but do not show, 3 = to be created/updated, 4 = to be deleted',
svg BLOB NOT NULL,
PRIMARY KEY (mol_id)
) ENGINE=MYISAM CHARACTER SET $mysql_charset COLLATE $mysql_collation COMMENT='Housekeeping for 2D depiction';";
if ($create_table == 1) {
  print "$createcmd\n";
}

# Now the tables have been (optionally) created, if necessary

# Then, get the next available mol_id number  ==> always 1 in this script
$entry = 0;
$counter  = 0;
$li       = 0;
$buf      = '';
$mol      = '';
$txt      = '';
$lbl      = '';
$ct       = 1;
$badmols  = 0;
$internal_name = "";


# process the input file line by line

while ($line = <SDF>) {
  $line =~ s/\r//g;      # remove carriage return characters (DOS/Win)
  if ((substr($line,0,4) eq '$$$$') || eof(SDF)) {
    $counter ++;
    $entry ++;
    $xentry = "XXXXn_" . $entry . "_nXXXX";   # use placeholder for mol_id so that it can be easily replaced
    if ($verbose > 1) { 
      print "-- adding entry $entry\n"; 
    } else {
      if ((($counter % 1000) == 0) && ($verbose > 0)) { print "-- $counter records processed\n"; }
    }
    if ($ct == 1) {
      $mol = $buf;
      $txt = "";
    } else {
      $txt = $buf;
      $txt =~ s/\;$/\;\ /g;    # append a blank space to trailing semicolons
    }
    $internal_name = "";
    if (valid_mol($mol) == 1) {
      $mol =~ s/\"/\\\"/g;        # escape any quote characters
      if (index($mol,"M  END") < 0) {
        $mol = $mol . "M  END\n"; # some SD files are lacking the "M  END" 
      }	
      insert_mol($mol);
      insert_data($txt,$mol);
      insert_molstat_molfgb_molcfp($mol);
      insert_svg($mol,$mol2svgopt,$svg_scalingfactor);
      print "-- XXXXn++\n";
    } else {
      $counter --;
      $entry --;
      $xentry = "XXXXn_" . $entry . "_nXXXX";      
      $badmols ++;
    }
    $buf = "";
    $txt = "";
    $mol = "";
    $ct  = 1;
  } else {
    if (substr($line,0,1) eq '>') { 
      if ($ct == 1) {
        $mol = $buf;
        $buf = $line;
      }
      $ct = 0; 
    }
    $buf = $buf . $line;
  }
}        # end while ($line....

if ($verbose > 0) {
  print "-- ==============================================================================\n";
  print "-- $counter records processed in total\n";
  print "-- $badmols records ignored\n\n";
}

#===================== subroutines =======================================

sub valid_mol() {
  $testmol = shift;
  $zerolines = 0;
  @xyzline = split(/\n/, $testmol);
  $internal_name = $xyzline[0];
  chomp($internal_name);
  for ($i = 0; $i <= $#xyzline; $i++) {
    $testline = $xyzline[$i];
    $testline =~ s/\ +/:/g;
    @xyz = split(/:/, $testline);
    $xval = $xyz[1];
    $yval = $xyz[2];
    $zval = $xyz[3];
    if ((index($xval,"0.0000") >= 0) && (index($yval,"0.0000") >= 0) && 
       (index($zval,"0.0000") >= 0)) { $zerolines ++; }
  }
  if ($chkzero == 0) { $zerolines = 0; }  # disable check if configured
  if ($zerolines > 1) {
    return 0;
  } else {
    return 1;
  }
}

sub insert_mol() {
  $molecule = shift;
  if ($tweakmolfiles eq "y") {
    if ($ostype eq 2) {
	  $molecule = filterthroughcmd2($molecule,"$CHECKMOL -m -");
	} else {
	  $molecule = filterthroughcmd($molecule,"$CHECKMOL -m -");
	}
  }
  $molecule =~ s/\m//g;      # clean the molfile from carriage-returns
  $molecule =~ s/\"/\'\'/g;  # clean the molfile from double quotes
  $molecule =~ s/\;+\n/\n/g;   # clean the molfile from trailing semicolons
  print "INSERT INTO $molstructable VALUES ( $xentry, \"$molecule\" ) ;\n";
  $mol = $molecule;
}

sub insert_data() {
  $data = shift;
  $molfile = shift;
  my $mfmw = get_mfmw($molfile);
  my @amfmw = split (/:/, $mfmw);
  my $mf = $amfmw[0];
  chomp($mf);
  my $mw = $amfmw[1];
  chomp($mw);
  $databuf  = "";
  @rec = split (/\n/, $data);
  for ($i = 0; $i <= $#rec+1; $i++) {
    $element  = $rec[$i];
    #$element =~ s/\n//g;
    chomp($element);
    $element =~ s/\ +$//g;
    $lblchars = 0;
    if (substr($element,0,1) eq '>') {
      @lblrec  = split(/\</, $element);
      $lblname = $lblrec[1];
      @lblrec  = split(/\>/, $lblname);
      $lblname = $lblrec[0];
    } else {
      if (length($element) > 0) {
        $databuf = $databuf . $element . "\n";  # add new-line
      } else {
        for ($j = 0; $j <= $#afield; $j++) {
          $knownlbl = $afield[$j][0];
          if ($lblname eq $knownlbl) { 
            chomp($databuf);   # remove last new-line
            $afield[$j][3] = $databuf;
          }
        }
        $databuf = "";
      }
    }
  }
  # create lists of field names and values
  $fieldlist = "mol_id";
  $valuelist = "\"$xentry\"";
  if ($found_mol_name == 0) {
    #$insertcmd = $insertcmd . ", \"$internal_name\"";
    $fieldlist .= "mol_name";
    $internal_name =~ s/\"/\\\"/g;
    $valuelist .= ", \"$internal_name\"";
  }  
  # handle auto_mol_ fields
  if ($found_auto_mol_formula == 1) {
    $fieldlist .= ", auto_mol_formula";
    $valuelist .= ", \"$mf\"";
  }  
  if ($found_auto_mol_fw == 1) {
    $fieldlist .= ", auto_mol_fw";
    $valuelist .= ", \"$mw\"";
  }  
  if ($found_auto_mol_inchikey == 1) {
    if ($ostype eq 2) {
      $inchi  = filterthroughcmd2($molfile,"$INCHI $INCHI_OPT");
    } else {
      $inchi  = filterthroughcmd($molfile,"$INCHI $INCHI_OPT");
    }
    @inchi_arr = split(/InChIKey=/,$inchi);
    $inchi = $inchi_arr[1];
    @inchi_arr = split(/\n/,$inchi);
    $inchikey = $inchi_arr[0];
    chomp($inchikey);
    if (length($inchikey) >= 25) {   # should be 27
      $fieldlist .= ", auto_mol_inchikey";
      $valuelist .= ", \"$inchikey\"";
    }
  }  
  for ($j = 0; $j <= $#afield; $j++) {
    $field = $afield[$j][1];
    $item = $afield[$j][3];
    $item =~ s/\\/\\\\/g;   # first, escape any backslashes
    $item =~ s/\"/\\\"/g;   # escape any double quotes
    #$insertcmd = $insertcmd . ",\n  \"$item\"";
    $fieldlist .= ", $field";
    $valuelist .= ", \"$item\"";
  }
  $insertcmd = "INSERT INTO $moldatatable ( $fieldlist ) VALUES ( $valuelist )"; 
  for ($j = 0; $j <= $#afield; $j++) {
    $afield[($j)][3] = "";
  }
  print "$insertcmd ;\n";
}

sub insert_molstat_molfgb_molcfp() {
  $molecule = shift;
  if ($ostype eq 2) {
    $molstatfgbhfp  = filterthroughcmd2($molecule,"$CHECKMOL -aXbH -");   # must be version 0.4 or higher
  } else {
    $molstatfgbhfp  = filterthroughcmd($molecule,"$CHECKMOL -aXbH -");   # must be version 0.4 or higher
  }
  @molstatfgbhfparray = split (/\n/, $molstatfgbhfp);
  $molstat  = $molstatfgbhfparray[0];
  $molfgb   = $molstatfgbhfparray[1];
  $molhfp   = $molstatfgbhfparray[2];
  if ((index($molstat,"unknown") < 0) && (index($molstat,"invalid") < 0)) {
    chomp($molstat);
    print "INSERT INTO $molstattable VALUES ( $xentry, $molstat ) ;\n";
    chomp($molfgb);
    $molfgb =~ s/\;/\,/g;
    print "INSERT INTO $molfgbtable VALUES ($xentry, $molfgb ) ;\n";
    $moldfp = "";
    for ($k = 0; $k < $n_dict; $k++) {
      $dict = @fpstruc[$k];
      $cand = $molecule . "\n" . '$$$$' ."\n" . $dict;
      $cand =~ s/\$/\\\$/g;
      if ($ostype eq 2) {
  	    $dfpstr = filterthroughcmd2($cand,"$MATCHMOL -F -");
	  } else {
	    $dfpstr = filterthroughcmd($cand,"$MATCHMOL -F -");
	  }
      chomp($dfpstr);
      if ($k > 0) { $moldfp .= ","; }
      $moldfp .= " " . $dfpstr;
    }
    chomp($molhfp);
    $molhfp =~ s/\;/\,/g;
    print "INSERT INTO $molcfptable VALUES ($xentry, $moldfp, $molhfp);\n";
  }
}

sub tweak_svg {
  my $testsvg = shift;
  my @svgline = split(/\n/, $testsvg);
  my $xmaxval = "";
  my $yminval = "";
  my $ytrval = "";
  my $scaling = $svg_scalingfactor;
  for ($i = 0; $i <= $#svgline; $i++) {
    my $testline = $svgline[$i];
    chomp($testline);
    if ((index($testline,"\<!-- max_X:") >= 0) && (index($testline,"--\>") >= 0)) {
      my @xline = split(/:/, $testline);
      $xmaxval = $xline[1];
      $xmaxval =~ s/--\>//g;
      chomp($xmaxval);
      $xmaxval =~ s/\ +//g;
      if ($keep_xydata == 0) { $svgline[$i] = ""; }
    }
    if ((index($testline,"\<!-- max_Y:") >= 0) && (index($testline,"--\>") >= 0)) {
      my @yline = split(/:/, $testline);
      $ymaxval = $yline[1];
      $ymaxval =~ s/--\>//g;
      chomp($ymaxval);
      if ($keep_xydata == 0) { $svgline[$i] = ""; }
    }
    if ((index($testline,"\<!-- min_Y:") >= 0) && (index($testline,"--\>") >= 0)) {
      my @yline = split(/:/, $testline);
      $yminval = $yline[1];
      $yminval =~ s/--\>//g;
      chomp($yminval);
      if ($keep_xydata == 0) { $svgline[$i] = ""; }
    }
    if ((index($testline,"\<!-- yshift:") >= 0) && (index($testline,"--\>") >= 0)) {
      my @ytrline = split(/:/, $testline);
      $ytrval = $ytrline[1];
      $ytrval =~ s/--\>//g;
      chomp($ytrval);
      if ($keep_xydata == 0) { $svgline[$i] = ""; }
    }
    if ((index($testline,"\<!-- found XY values for adjusting") >= 0) && (index($testline,"--\>") >= 0)) {
      if ($keep_xydata == 0) { $svgline[$i] = ""; }
    }
  }  # for
  if ((length($xmaxval) > 0) && (length($ymaxval) > 0) && (length($yminval) > 0) && (length($ytrval) > 0)) {
    my $ymaxtotal   = $ymaxval + $ytrval;
    my $ymintotal   = $yminval + $ytrval;
    my $ydiff       = $ymaxtotal - $ymintotal;
    my $ydiffscaled = $ydiff * $scaling;
    my $xmaxscaled  = $xmaxval * $scaling;
    my $ymaxscaled  = $ymaxtotal * $scaling;
    my $yminscaled  = $ymintotal * $scaling;
    $svgline[1] = "<svg width=\"$xmaxscaled\" height=\"$ydiffscaled\" viewbox=\"0 $ymintotal $xmaxval $ydiff\" xmlns=\"http://www.w3.org/2000/svg\">";
  }
  my $twsvg = join("\n",@svgline);
  $twsvg =~ s/\n\n\n//g;
  return $twsvg;
}


sub insert_svg() {
  $structure = shift;
  $mopt = shift;
  $sf = shift;
  $structure =~ s/\"/\\\"/g;
  $structure =~ s/\$/\\\$/g;
  if (index($structure,"M  END") < 0) { $structure = $structure . "M  END\n"; }	
  if ($ostype eq 2) {
    $molsvg = filterthroughcmd2($structure,"$MOL2SVG $mopt - ");
  } else {
    $molsvg = filterthroughcmd($structure,"$MOL2SVG $mopt - ");
  }
  #print "$molsvg\n";
  #$newsvg = tweak_svg($molsvg);   # needed only for mol2svg v0.3x
  $newsvg = $molsvg;   # for mol2svg v0.4 or higher
  $newsvg =~ s/\"/\\\"/g;
  if ($keep_comments == 0) {
    $newsvg =~ s/(\<!--.*?--\>)//g;
  }
  $newsvg =~ s/\n\n+/\n/g;  # remove excessive newlines
  #print "$newsvg\n";
  print "INSERT INTO $pic2dtable VALUES ( $xentry, 1, 3, \"${newsvg}\" );\n";
}


sub filterthroughcmd {
  $input   = shift;
  $cmd     = shift;
  open(FHSUB, "echo \"$input\"|$cmd 2>&1 |");   # stderr must be redirected to stdout
  $res      = "";                               # because the Ghostscript "bbox" device
  while($line = <FHSUB>) {                      # writes to stderr
    $res = $res . $line;
  }
  return $res;
}

sub filterthroughcmd2 {                         # workaround for Windows 
  $input   = shift;
  $cmd     = shift;
  ($tmpfh, $tmpfilename) = tempfile(UNLINK => 1);
  $input =~ s/\\\$/\$/g;
  $input =~ s/\r//g;
  $input =~ s/\n/\r\n/g;
  print $tmpfh "$input\n";
  #open(FHSUB, "type $tmpfilename |$cmd 2>&1 |");   # stderr must be redirected to stdout
  open(FHSUB, "$cmd < $tmpfilename 2>&1 |");   # stderr must be redirected to stdout
  $res      = "";                               # because the Ghostscript "bbox" device
  while($line = <FHSUB>) {                      # writes to stderr
    $res = $res . $line;
  }
  close $tmpfh;
  return $res;
}

sub ltrim() {
  $subline1 = shift;
  $subline1 =~ s/^\ +//g;
  return $subline1;
}

sub rtrim() {
  $subline2 = shift;
  $subline2 =~ s/\ +$//g;
  return $subline2;
}

sub getostype() {
  $os  = "";
  $osresult = 1;
  $os  = uc($ENV{OS});
  if ($os eq "") { $os = uc($ENV{OSTYPE}); }
  if (index($os,"WINDOWS")>=0) {
    $osresult = 2;
  }
  return $osresult;
}

sub get_mfmw() {
  my $inmol = shift;
  #print "$inmol\n";
  my @marr = split (/\n/, $inmol );
  my $mfmwline = $marr[2];
  #print ".... $mfmwline\n";
  my @mfmwa1 = split (/MW=/,$mfmwline);
  my $mw = $mfmwa1[1];
  chomp($mw);
  #print "    === MW = $mw\n";
  my $mf1 = $mfmwa1[0];
  my @mfmwa2 = split (/MF=/,$mf1);
  my $mf = $mfmwa2[1];
  $retval = "$mf" . ":" . "$mw";
  return $retval;
}
