#!/usr/bin/perl
#
# sql2moldb.pl   Norbert Haider, University of Vienna, 2015-2021
#                norbert.haider@univie.ac.at
#
# This script is part of the MolDB6 package. Last change: 2021-02-28
#
# This script reads a pre-processed SQL file that was written by the
# sdf2sql.pl script. It adjusts the mol_id number and appends all
# records to a MolDB6 data collection.
#
# ===========================================================================
#
# Purpose: setting up a web-based, fully searchable molecular structure 
# database (MolDB6); for a more detailed description see
# https://homepage.univie.ac.at/norbert.haider/cheminf/moldb.html
#
# IMPORTANT: by default, all data are _appended_, no redundancy checks for 
# already existing records are made.
#
# This software was developed and tested with the SDF files from the 
# PubChem database, which are freely available at
# ftp://ftp.ncbi.nlm.nih.gov/pubchem/Compound/CURRENT-Full/SDF/

$verbose          = 1;   # 0, 1 or 2
$askuser          = 0;   # 0 or 1, set to 0 for skipping the confirmation
$append           = 1;   # 0 or 1 (if 0, all existing data will be erased)

if ($#ARGV < 0) {
  print "Usage: sql2moldb.pl <inputfile>\n";
  exit;
}

$infile = $ARGV[0];

use DBI();
$ostype = getostype();
if ($ostype eq 2) { use File::Temp qw/ tempfile tempdir /; }

$configfile = "../moldb6.conf";

$return     = do $configfile;
if (!defined $return) {
  die("ERROR: cannot read configuration file $configfile!\n");
}	


$user     = $rw_user;    # from configuration file
$password = $rw_password;

$mysql_charset   = "latin1";          # default
$mysql_collation = "latin1_swedish_ci";
$isnewdb = 1;

if ($charset eq "latin2") {           # from configuration file
  $mysql_charset   = "latin2";
  $mysql_collation = "latin2_general_ci";
}
if ($charset eq "utf8") {
  $mysql_charset   = "utf8";
  $mysql_collation = "utf8_unicode_ci";
}

$db_id                   = 1;  # default data collection number
$db_access               = 1;  # default access mode (0 = hidden, 1 = read-only, 2 = read/write)
$bitmapfile_digits       = 8;  # default values for bitmap filenames (e.g., 00000123.png)
$bitmapfile_subdirdigits = 4;  # default value for subdirectory names (e.g., 0000/00000123.png)
$db_name                 = "";
$flags                   = 0;  # 32bit integer (unsigned); currently unused

$deffile = "sdf2moldb.def";
$found_mol_id   = 0;
$found_mol_name = 0;
open (DEF, "<$deffile") || die("ERROR: cannot open definition file $deffile!");
$nfields = 0;
while ($line = <DEF>) {
  chomp($line);
  @valid = split (/#/, $line);  # ignore everything behind the first pound sign
  $line  = $valid[0];
  $line  = ltrim($line);
  $line  = rtrim($line);
  if ((index($line,'sdfilename') == 0) && (index($line,'=') >= 10)) {
    @defrec = split (/=/, $line);
    $sdfile = $defrec[1];
    $sdfile = ltrim($sdfile);
  }
  if ((index($line,'db_id') == 0) && (index($line,'=') >= 5)) {
    @defrec = split (/=/, $line);
    $db_id = $defrec[1];
    $db_id = ltrim($db_id);
  }
  if ((index($line,'db_name') == 0) && (index($line,'=') >= 7)) {
    @defrec = split (/=/, $line);
    $db_name = $defrec[1];
    $db_name = ltrim($db_name);
    $db_name =~ s/\"//g;
  }
  if ((index($line,'db_description') == 0) && (index($line,'=') >= 14)) {
    @defrec = split (/=/, $line);
    $db_description = $defrec[1];
    $db_description = ltrim($db_description);
    $db_description =~ s/\"//g;
  }
  if ((index($line,'db_access') == 0) && (index($line,'=') >= 9)) {
    @defrec = split (/=/, $line);
    $tmp_access = $defrec[1];
    $tmp_access = ltrim($tmp_access);
    if ($tmp_access == 0) { $db_access = 0; }
    if ($tmp_access == 1) { $db_access = 1; }
    if ($tmp_access == 2) { $db_access = 2; }
    if ($tmp_access == 3) { $db_access = 3; }    
  }
  if ((index($line,'bitmapfile_digits') == 0) && (index($line,'=') >= 17)) {
    @defrec = split (/=/, $line);
    $tmp_digits = $defrec[1];
    $tmp_digits = ltrim($tmp_digits);
    if (($tmp_digits > 0) && ($tmp_digits <= 32)) { $bitmapfile_digits = $tmp_digits; }
  }
  if ((index($line,'bitmapfile_subdirdigits') == 0) && (index($line,'=') >= 23)) {
    @defrec = split (/=/, $line);
    $tmp_subdirdigits = $defrec[1];
    $tmp_subdirdigits = ltrim($tmp_subdirdigits);
    if (($tmp_subdirdigits >= 0) && ($tmp_subdirdigits <= $bitmapfile_digits)) { $bitmapfile_subdirdigits = $tmp_subdirdigits; }
  }
  if ((index($line,'flags') == 0) && (index($line,'=') >= 4)) {
    @defrec = split (/=/, $line);
    $tmp_flags = $defrec[1];
    $tmp_flags = ltrim($tmp_flags);
    chomp($tmp_flags);
    if (($tmp_flags + 0 eq $tmp_flags ) && ( $tmp_flags >= 0 ) && ( $tmp_flags < 2147483648 ) ) {
      $flags = $tmp_flags;
    } else {
      $flags = 0;
    }
  }
  
  $lpos = index($line,':');
  $rpos = rindex($line,':');
  if (($lpos >= 1) && ($rpos >= 3) && ($rpos > $lpos)) {
    # this should be a definition line
    @defrec = split (/:/, $line);
    $sdf_label   = $defrec[0];
    $mysql_label = $defrec[1];
    $mysql_type  = $defrec[2];
    $html_label  = $defrec[3];
    $html_format = $defrec[4];
    if (index($mysql_label,"auto_mol_") != 0) {
      @afield[($nfields)] = [ $sdf_label, $mysql_label, $mysql_type, $html_label, $html_format, "" ];
      $nfields++;
    }
    if ($mysql_label eq "mol_name") { $found_mol_name = 1; }
    if ($mysql_label eq "mol_id") { $found_mol_id = 1; print "$line\n"; }
  }
}

if ($verbose > 0) { 
  print "\n-- Adding data to MolDB6 database '$database', using configuration\n";
  print "-- file '$configfile'.\n\n";
  print "-- Your SDF input file must have the same format as the one which\n";
  print "-- has been used previously for analysis with the sdfcheck.pl script:\n";
  print "-- $sdfile\n\n";
  if ($append eq 0) {
    print "-- WARNING: all existing data in this collection will be erased!\n\n";
  }
  	
}

if ($found_mol_id > 0) {
  print STDERR "\nERROR: your definition file $deffile contains a field named\n";
  print STDERR "'mol_id'! This label must not be used, as it is automatically created\n";
  print STDERR "by the system. Please edit $deffile and change 'mol_id' into some\n";
  print STDERR "other name.\n";
  exit;
}

if ($found_mol_name == 0) {
  print "\n-- WARNING: your definition file $deffile does not contain a field\n";
  print "-- named 'mol_name'. It is highly recommended to rename the most descriptive\n";
  print "-- field into 'mol_name', please read the instructions in $deffile!\n";
  print "-- Otherwise, a field 'mol_name' will be automatically created, but it\n";
  print "-- might remain empty during the SDF import operation.\n\n";
  $askuser = 1;
}

if ($askuser > 0) {
  print "Do you really want to continue (y/n)? ";
  chomp($word = <STDIN>);
  if (lc($word) eq 'y') {
    if ($verbose > 0) { 
      print "-- OK. This will take some time, please be patient.\n";
    }
  } else {
    print "-- aborting...\n";
    exit;
  }
}

if ($verbose > 1) { 
  print "-- using SD file $infile\n"; 
  print "-- found the following data field definitions:\n";
  for ($i = 0; $i <= $#afield; $i++) {
    $l1  = $afield[$i][0];
    $l2  = $afield[$i][1];
    $l3  = $afield[$i][2];
    $l4  = $afield[$i][3];
    $l5  = $afield[$i][4];
    while (length($l1) < 25) { $l1 = $l1 . " "; }
    while (length($l2) < 25) { $l2 = $l2 . " "; }
    while (length($l3) < 25) { $l3 = $l3 . " "; }
    while (length($l4) < 5) { $l4 = $l4 . " "; }
    print "-- $l1   $l2  $l3 $l4 $l5\n";
  } 
}

if ($verbose > 0) { 
  print "\n-- these settings are used:\n";
  print "--  MySQL database:           $database\n";
  print "--  hostname:                 $hostname\n";
  print "--  admin user:               $user\n";
  print "--  SDF file:                 $infile\n";
  print "--  data collection number:   $db_id\n";
  print "--  data collection name:     $db_name\n";
  print "--  description:              $db_description\n";
  print "--  access mode (0,1,2,3):    $db_access\n";
  print "--  bitmap directory:         $bitmapdir\n";
  print "--  bitmap filename digits:   $bitmapfile_digits\n";
  print "--  subdirectory name digits: $bitmapfile_subdirdigits\n\n";
  print "--  flags:                    $flags\n\n";  
}


if ((!defined $user) || ($user eq "")) {
  die("ERROR: no username specified!\n");
}



# open the MySQL database and check if data collection exists already

$dbh = DBI->connect("DBI:mysql:database=$database;host=$hostname",
                    $user, $password,
                    { RaiseError => 1}
                    ) || die("ERROR: database connection failed: $DBI::errstr");
$qstr = "SET NAMES " . $mysql_charset;
$dbh->do($qstr);
$qstr = "SET SESSION sql_mode = '';";
$dbh->do($qstr);

#print "USE $database ;\n $qstr ;\n";

$qstr = "SELECT db_id, name FROM $metatable WHERE db_id = $db_id";
$sth0 = $dbh->prepare($qstr);
$sth0->execute();
$n = 0;
while ($ref0 = $sth0->fetchrow_hashref()) {
  $n++;
  $id   = $ref0->{'db_id'};
  $name = $ref0->{'name'};
}

if ($n > 0) {
  if ($name eq $db_name) {
    if ($verbose > 0) {
      print "-- INFORMATION: adding data to an already existing collection\n";
    }
    $isnewdb = 0;
  } else {
    print STDERR "ERROR: a data collection with this number, but with a different\n";
    print STDERR "name ($name) exists already: check your settings in $deffile!\n";
    exit;
  }	
} else {
  if ($verbose > 0) {
    print "-- INFORMATION: this is a new data collection\n";	
  }
  $isnewdb = 1;
}

$dbprefix      = $prefix . "db" . $db_id . "_";
$molstructable = $dbprefix . $molstrucsuffix;
$moldatatable  = $dbprefix . $moldatasuffix;
$molstattable  = $dbprefix . $molstatsuffix;
$molfgbtable   = $dbprefix . $molfgbsuffix;
$molcfptable   = $dbprefix . $molcfpsuffix;
$pic2dtable    = $dbprefix . $pic2dsuffix;


open (SQL, "<$infile") || die("ERROR: cannot open input file $infile!");

if ($isnewdb == 1) {
  $insertcmd = "INSERT INTO $metatable VALUES ( $db_id, 1, $db_access, \"$db_name\", ";
  $insertcmd .= " \"$db_description\", \"F\", 0, $bitmapfile_digits, $bitmapfile_subdirdigits, \"\", \"$flags\")";
  $dbh->do($insertcmd);
}


# Next, disable use of memory-based tables
$updstr = "UPDATE $metatable SET memstatus = 0 WHERE db_id = $db_id";
$dbh->do($updstr);	
#print "$updstr ;\n";

# Remove any existing data if $append is set to 0

if (($append eq 0) && ($isnewdb eq 0)) {
  if ($verbose > 0) {
    print "-- erasing existing data... \n";
  }
  $delcmd = "TRUNCATE TABLE $molstructable; ";
  $dbh->do($delcmd);
  $delcmd = "TRUNCATE TABLE $moldatatable; ";
  $dbh->do($delcmd);
  $delcmd = "TRUNCATE TABLE $molstattable; ";
  $dbh->do($delcmd);
  $delcmd = "TRUNCATE TABLE $molfgbtable; ";
  $dbh->do($delcmd);
  $delcmd = "TRUNCATE TABLE $molcfptable; ";
  $dbh->do($delcmd);
  $delcmd = "TRUNCATE TABLE $pic2dtable; ";
  $dbh->do($delcmd);
}

# Then, get the next available mol_id number

$entry = 0;

if ($isnewdb == 0) {
  $sth0  = $dbh->prepare("SELECT mol_id FROM $moldatatable ORDER BY mol_id DESC LIMIT 0,1 ");
  $sth0->execute();
  while ($ref0 = $sth0->fetchrow_hashref()) {
    $entry = $ref0->{'mol_id'};
  }
  if ($verbose > 1) { print "-- number of molecules already in the database: $entry \n"; }
}

$entry++;


$counter  = 0;
$li       = 0;
$buf      = '';
$mol      = '';
$txt      = '';
$lbl      = '';
$ct       = 1;
$badmols  = 0;
$internal_name = "";


# process the input file line by line

while ($line = <SQL>) {
  $line =~ s/\r//g;      # remove carriage return characters (DOS/Win)
  #print "$line";
  if ((substr($line,0,9) eq '-- XXXXc-') && ($isnewdb == 1)) { 
    print "-- ERROR: this is a new data collection, but the CREATE TABLE statements are missing in $infile!\n";
    $delcmd = "DELETE FROM $metatable WHERE db_id = $db_id";
    $dbh->do($delcmd);
    $dbh->disconnect();
    exit;
  }
  if (substr($line,0,10) eq '-- XXXXn++') { 
    $entry++ ; 
    $counter ++;
    if ((($counter % 1000) == 0) && ($verbose > 0)) { print "-- $counter records processed\n"; }
  } else {
    $buf = $buf . $line;
    if ((substr($line,-2,1) eq ';') || eof(SQL)) {
      $buf =~ s/XXXXn_[0-9]*_nXXXX/$entry/g;
      #print "$buf \n";
      $dbh->do($buf);
      $buf = "";
    }
  }
}        # end while ($line....

$dbh->disconnect();

if ($verbose > 0) {
  print "-- ==============================================================================\n";
  print "-- $counter records processed in total\n";
  print "-- $badmols records ignored\n\n";
}

#===================== subroutines =======================================


sub filterthroughcmd {
  $input   = shift;
  $cmd     = shift;
  open(FHSUB, "echo \"$input\"|$cmd 2>&1 |");   # stderr must be redirected to stdout
  $res      = "";                               # because the Ghostscript "bbox" device
  while($line = <FHSUB>) {                      # writes to stderr
    $res = $res . $line;
  }
  return $res;
}

sub filterthroughcmd2 {                         # workaround for Windows 
  $input   = shift;
  $cmd     = shift;
  ($tmpfh, $tmpfilename) = tempfile(UNLINK => 1);
  $input =~ s/\\\$/\$/g;
  $input =~ s/\r//g;
  $input =~ s/\n/\r\n/g;
  print $tmpfh "$input\n";
  #open(FHSUB, "type $tmpfilename |$cmd 2>&1 |");   # stderr must be redirected to stdout
  open(FHSUB, "$cmd < $tmpfilename 2>&1 |");   # stderr must be redirected to stdout
  $res      = "";                               # because the Ghostscript "bbox" device
  while($line = <FHSUB>) {                      # writes to stderr
    $res = $res . $line;
  }
  close $tmpfh;
  return $res;
}

sub ltrim() {
  $subline1 = shift;
  $subline1 =~ s/^\ +//g;
  return $subline1;
}

sub rtrim() {
  $subline2 = shift;
  $subline2 =~ s/\ +$//g;
  return $subline2;
}

sub getostype() {
  $os  = "";
  $osresult = 1;
  $os  = uc($ENV{OS});
  if ($os eq "") { $os = uc($ENV{OSTYPE}); }
  if (index($os,"WINDOWS")>=0) {
    $osresult = 2;
  }
  return $osresult;
}

