#!/usr/bin/perl
#
# updatebitmap.pl   Norbert Haider, University of Vienna, 2009-2018
#                   norbert.haider@univie.ac.at
#
# This script is part of the MolDB6 package. Last change: 2018-01-05
#
# Example script which (re-)generates the bitmap files for 2D depiction
# of the molecules in the MolDB6 database: this gives better performance 
# and a better quality of displayed structures than using the JME
# applet in "depict" mode for each hit structure. Usually, this script
# is run as a cron job.

# Requirements: mol2ps and Ghostscript must be installed.

use DBI();

$check      = 3;   # 0 = no checks at all
                   # 1 = check for missing pic2d tables and records
                   # 2 = as 1, plus check for orphan files (to be deleted)
                   # 3 = as 2, plus check for missing files

$verbose    = 1;   # 0 = silent operation, 
                   # 1 = report each data collection + principal action
                   # 2 = report each molecule

$reallydelete = 1; # 0 = only rename files which are no longer needed
                   # 1 = really delete these files
                   
#$MOL2PS       = "mol2ps";   # read from configuration file
##$GHOSTSCRIPT  = "/usr/bin/gs";  # for Linux
#$GHOSTSCRIPT  = "gswin32c";   # for Windows
$enablerxnbitmaps = "y";  # for backward compatibility

$configfile = "../moldb6.conf";
#$configfile = "/data/moldb/moldb6/moldb6.conf";  # better use absolute path when
                                                   # running this script via cron

if ($#ARGV >= 0) {
  for ($a = 0; $a <= $#ARGV; $a++) {
    $aline = $ARGV[$a];
    if (($aline eq '-h') || ($aline eq '-?') || ($aline eq '--help')) {
      print "Usag: updatebitmap.pl [--check=<n>]    (with <n> as 0, 1, 2, 3, or full)\n";
      exit;
    }
    if (index($aline,"--check=") >= 0) {
      $checkval = getval($aline);
      if ($checkval eq "0") { $check = 0; }
      if ($checkval eq "1") { $check = 1; }
      if ($checkval eq "2") { $check = 2; }
      if ($checkval eq "3") { $check = 3; }
      if ($checkval eq "full") { $check = 3; }
      
    }
  }
}


$ostype = getostype();
if ($ostype eq 2) { use File::Temp qw/ tempfile tempdir /; }

$return     = do $configfile;
if (!defined $return) {
  die("ERROR: cannot read configuration file $configfile!\n");
}	

$user     = $rw_user;    # from configuration file
$password = $rw_password;

$mysql_charset   = "latin1";          # default
$mysql_collation = "latin1_swedish_ci";

if ($charset eq "latin2") {           # from configuration file
  $mysql_charset   = "latin2";
  $mysql_collation = "latin2_general_ci";
}
if ($charset eq "utf8") {
  $mysql_charset   = "utf8";
  $mysql_collation = "utf8_unicode_ci";
}

if ((defined $bitmapdir) && ($bitmapdir ne "")) {
  if ( ! -d $bitmapdir) {
    mkdir $bitmapdir || die("ERROR: cannot create directory $bitmapdir !");
    if ($verbose > 0) {
      print "created directory: $bitmapdir\n";
    }
  }
  if ( ! -W $bitmapdir) {
    die("ERROR: cannot write to directory $bitmapdir !");
  }
  # check for mol2ps
  $return = `$MOL2PS -v`;
  if (index($return,"Usage:") < 0) {
    die("ERROR: could not find 'mol2ps', make sure it is installed and in your search path");	
  }  
  if (index($return,"mol2ps version 0.1") >= 0) {
    print "WARNING: your mol2ps is older than version 0.2 and does not support reaction files!\n";
    $enablerxnbitmaps = "n";
  }  
  
  # check for Ghostscript
  $return = `$GHOSTSCRIPT -v`;
  if (index($return,"Ghostscript") < 0) {
    die("ERROR: could not find ghostscript, make sure it is installed and in your search path");	
  }  
} else {
  die("ERROR: bitmap directory is not defined in your configuration file!");
}


$dbh = DBI->connect("DBI:mysql:database=$database;host=$hostname",
                    $user, $password,
                    {'RaiseError' => 1});
$qstr = "SET SESSION sql_mode = '';";
$dbh->do($qstr);

# read moldb_meta table and find out which data collections are to be processed
$ndb = 0;
@db = [""];
#$sth0 = $dbh->prepare("SELECT db_id FROM $metatable WHERE type = 1 ORDER BY db_id");
$wherestr = "WHERE type = 1";
if ($enablerxnbitmaps == "y") { $wherestr = ""; }
$sth0 = $dbh->prepare("SELECT db_id, type FROM $metatable ORDER BY db_id");
$sth0->execute();
while ($ref0 = $sth0->fetchrow_hashref()) {
  $db_id   = $ref0->{'db_id'};
  $db_type = $ref0->{'type'};
  $ndb++;
  @db[($ndb-1)] = [ $db_id, $db_type ];
}
$sth0->finish;

$badmols = 0;
$counter = 0;
$delcounter = 0;

for ($idb = 0; $idb < $ndb; $idb++) {
  $dbnum  = $db[$idb][0];
  $dbtype = $db[$idb][1];
  $dbprefix = $prefix . "db" . $dbnum . "_";
  $pic2dtable    = $dbprefix . $pic2dsuffix;
  if ($dbtype == 1) {
    $idname = "mol_id";
    $structable = $dbprefix . $molstrucsuffix;
    $options = $mol2psopt;
    $scale = $scalingfactor;
  }
  if ($dbtype == 2) {
    $idname = "rxn_id";
    $structable = $dbprefix . $rxnstrucsuffix;
    if ( (!defined $mol2psopt_rxn) || ($mol2psopt_rxn eq "")) {
      $options = "-R " . $mol2psopt;  
    } else {
      $options = $mol2psopt_rxn;
    }
    if ( (!defined $scalingfactor_rxn) || ($scalingfactor_rxn eq "")) {
      $scale = $scalingfactor;
    } else {
      $scale = $scalingfactor_rxn;
    }
  }
  # get number of digits and subdirdigits for PNG filenames of this data collection
  $sth0 = $dbh->prepare("SELECT digits, subdirdigits FROM $metatable WHERE db_id = $dbnum");
  $sth0->execute();
  while ($ref0    = $sth0->fetchrow_hashref()) {
    $digits       = $ref0->{'digits'};
    $subdirdigits = $ref0->{'subdirdigits'};
  }
  $sth0->finish;
  # check if the number of PNG filename digits is OK
  if (!defined $digits) {
    $digits = 8;
    if ($verbose > 0) {
      print "could not get number of digits, using default($digits)\n";
    }
  }
  # check if subdirectories are to be used
  if (!defined $subdirdigits) {
    $subdirdigits = 0;
    if ($verbose > 0) {
      print "could not get number of subdir digits, using default($subdirdigits)\n";
    }
  }
  
  if ($check > 0) {
    pic2d_check();
  }
  
  # read all pending molecules from mol/rxn-structable and pipe the MDL molfiles
  # through mol2ps and Ghostscript to produce the PNG graphics files
  #
  # first, get the number of rows and chop the whole operation into
  # suitable chunks
  if ($check > 2) { 
    $wherestatus = "WHERE (${pic2dtable}.status = 1) OR (${pic2dtable}.status = 3)";
  } else { 
    $wherestatus = "WHERE (${pic2dtable}.status = 3)"; 
  }
  $sth0 = $dbh->prepare("SELECT COUNT($idname) AS itemcount FROM $pic2dtable $wherestatus ");
  $sth0->execute();
  while ($ref0 = $sth0->fetchrow_hashref()) {
    $itemcount = $ref0->{'itemcount'};
  }
  $sth0->finish();
  if ($verbose > 0) {
    print "number of entries in data collection $dbnum to be processed: $itemcount \n";
  }
  $nchunks = int( (($itemcount + 99) / 100) );
  $li  = 0;
  $buf = "";
  $mol = "";
  $txt = "";
  $lbl = "";
  $ct  = 1;
  if ($itemcount > 0) {
    $dbbitmapdir = $bitmapdir . "/" . $dbnum;
    if ( ! -d $dbbitmapdir) {
      mkdir $dbbitmapdir || die("ERROR: cannot create directory $dbbitmapdir !");
      if ($verbose > 1) { print "created subdirectory: $dbbitmapdir\n"; }
    }
    if ( ! -W $dbbitmapdir) {
      die("ERROR: cannot write to directory $dbbitmapdir !");
    }
    # first round: check for missing files, if $check > 2
    if ($check > 2) {
      $last_id = 0;
      for ($ii = 0; $ii < $nchunks; $ii++) {
        $qstr = "SELECT $idname, status FROM $pic2dtable WHERE $idname > $last_id LIMIT 0, 100";
        $sth = $dbh->prepare($qstr);
        $sth->execute();
        $rows = $sth->rows();
        if ($rows > 0) {        
          while ($ref = $sth->fetchrow_hashref()) {
            $item_id   = $ref->{"$idname"};
            $last_id   = $item_id;
            $status    = $ref->{"status"};
            if ($status == 1) {  # only in this case, we do the check
              $fname = $item_id;
              while (length($fname) < $digits) { $fname = "0" . $fname; }
              $subdir = '';
              if ($subdirdigits > 0) {
                $subdir = substr($fname,0,$subdirdigits);
                $newdir = $dbbitmapdir . "/" . $subdir;
                $subdir = $subdir . "/";        
              }
              $fname = $dbbitmapdir . "/" . $subdir . $fname . ".png";
              if ( ! -f $fname) { 
                if ($verbose > 1) { print " missing file: $fname\n"; }
                $sth2 = $dbh->prepare("UPDATE $pic2dtable SET status = 3 WHERE $idname = $item_id");
                $sth2->execute();
                $sth2->finish();
              }
            }
          }  # while ...
        }  # if ($rows > 0)...
        $sth->finish;
      }  # for ...
    }
    
    for ($ii = 0; $ii < $nchunks; $ii++) {
      # no offset for LIMIT as the searched item (status) is permanently updated
      # (and non-zero offsets in large tables can result in _very_ slow queries!)
      $qstr = "SELECT ${structable}.$idname, ${structable}.struc FROM ";
      $qstr .= "$structable, $pic2dtable WHERE (status = 3) AND ";
      $qstr .= "(${structable}.$idname = ${pic2dtable}.$idname) LIMIT 0, 100";
      $sth = $dbh->prepare($qstr);
      $sth->execute();
      while ($ref = $sth->fetchrow_hashref()) {
        $item_id   = $ref->{"$idname"};
        $struc     = $ref->{'struc'};
        $counter ++;
        $isvalid = 0;
        if ( $dbtype == 1) { $isvalid = valid_mol($struc); }
        if ( $dbtype == 2) { $isvalid = valid_rxn($struc); }
        if ($isvalid eq 1) {
          $fname = $item_id;
          while (length($fname) < $digits) { $fname = "0" . $fname; }
          $subdir = '';
          if ($subdirdigits > 0) {
            $subdir = substr($fname,0,$subdirdigits);
            $newdir = $dbbitmapdir . "/" . $subdir;
            if ( ! -d $newdir) {
              mkdir $newdir || die("ERROR: cannot create directory $newdir !");
              if ($verbose > 1) { print "created subdirectory: $newdir\n"; }
            }
            if ( ! -W $newdir) {
              die("ERROR: cannot write to directory $newdir !");
            }
            $subdir = $subdir . "/";        
          }
          $fname = $dbbitmapdir . "/" . $subdir . $fname . ".png";

          # mark this id as "in progress" (30) in pic2dtable.status
          $sth2 = $dbh->prepare("UPDATE $pic2dtable SET status = 30 WHERE $idname = $item_id");
          $sth2->execute();
          $sth2->finish();
          process_struc($struc,$fname,$options,$scale);
          #if exists(fname)...
          if ( -f $fname) {
            # mark this id as "done" (1) in pic2dtable.status
            $sth2 = $dbh->prepare("UPDATE $pic2dtable SET status = 1 WHERE $idname = $item_id");
            $sth2->execute();
            $sth2->finish();
          }
        } else {
          $counter --;
          $badmols ++;
        }
      }                 # end while ($ref...
      $sth->finish;
    }                   # end "for" loop
  }   # if ...    

  #===================================check for bitmap files for deletion
  if ($check > 1) {
    $dbbitmapdir = $bitmapdir . "/" . $dbnum;
    if ($verbose > 0) {
      print " checking for orphan bitmap files in $dbbitmapdir...\n";
    }
    if ($ostype eq 1) { open(FIND,"find $dbbitmapdir -name \"*.png\" -print |"); }
    if ($ostype eq 2) { 
      $dbbitmapdir_win = $dbbitmapdir;
  	$dbbitmapdir_win =~ s/\//\\/g;
  	$fullpath = $dbbitmapdir_win . "\\\*.png";
  	open(FIND,"dir $fullpath /b /s |"); 
    }
    while ($fullname = <FIND>) {
      chomp($fullname);  
  	if ($ostype eq 2) { $fullname =~ s/\\/\//g; }
      $slashpos = rindex($fullname,"/");
      $fname = substr($fullname,($slashpos + 1));
      $item_id = $fname;
      $item_id =~ s/\.png//g;
      # check if this id exists in our data collection
      $itemcount = 1;
      $sth0 = $dbh->prepare("SELECT COUNT($idname) AS itemcount FROM $pic2dtable WHERE $idname = $item_id ");
      $sth0->execute();
      while ($ref0 = $sth0->fetchrow_hashref()) {
        $itemcount = $ref0->{'itemcount'};
      }
      $sth0->finish();
      if ($itemcount < 1) {
        if ($verbose > 1) {
          if ($reallydelete == 1) { 
            print " deleting file $fullname\n";
          } else {
            print " $fullname marked to be deleted\n";
          }
        }
        $delcounter++;
        if ($reallydelete == 1) { 
          (unlink $fullname) || print STDERR "ERROR: failed to delete $fullname\n";
        } else {
          $delname = $fullname . ".to_be_deleted";
          (rename $fullname, $delname) || print STDERR "ERROR: renaming failed for $fullname\n";
        }
      }
    }
  }
  #=================================end deletion

} # for ($idb ....


$dbh->disconnect();

if ($verbose > 0) {
  print "$counter records processed in total\n";
  print "$badmols records ignored\n";
  print "$delcounter bitmap file(s) deleted or marked to be deleted\n";
}

#============================================================

sub valid_mol() {
  $testmol = shift;
  $zerolines = 0;
  @xyzline = split(/\n/, $testmol);
  for ($i = 0; $i <= $#xyzline; $i++) {
    $testline = $xyzline[$i];
    $testline =~ s/\ +/:/g;
    @xyz = split(/:/, $testline);
    $xval = $xyz[1];
    $yval = $xyz[2];
    $zval = $xyz[3];
    if ((index($xval,"0.0000") >= 0) && (index($yval,"0.0000") >= 0) && 
       (index($zval,"0.0000") >= 0)) { $zerolines ++; }
  }
  if ($zerolines > 1) {
    return 0;
  } else {
    return 1;
  }
}

sub valid_rxn() {
  my $testrxn = shift;
  my $result = 0;
  if ((index($testrxn,'$RXN') == 0) && (index($testrxn,'$MOL') > 0) && (index($testrxn,'M  END') > 0)) {
    $result = 1;
  }
  return($result);
}

sub process_struc() {
  $structure = shift;
  $filename = shift;
  $mopt = shift;
  $sf = shift;
  $gsdevice = "pnggray";
  if (index($mopt,"--color=") >= 0) { $gsdevice = "png256"; }
  $structure =~ s/\"/\\\"/g;
  $structure =~ s/\$/\\\$/g;
  if (index($structure,"M  END") < 0) { $structure = $structure . "M  END\n"; }	

  if ($ostype eq 2) {
    $molps = filterthroughcmd2($structure,"$MOL2PS $mopt - ");
    $bb =  filterthroughcmd2($molps,"$GHOSTSCRIPT -q -sDEVICE=bbox -dNOPAUSE -dBATCH  -r300 -g500000x500000 - ");
  } else {
    $molps = filterthroughcmd($structure,"$MOL2PS $mopt - ");
    $bb =  filterthroughcmd($molps,"$GHOSTSCRIPT -q -sDEVICE=bbox -dNOPAUSE -dBATCH  -r300 -g500000x500000 - ");
  }
  @bbrec =   split(/\n/, $bb);
  $bblores = $bbrec[0];
  $bblores =~ s/%%BoundingBox://g;
  chomp($bblores);
  $bblores = ltrim($bblores);
  @bbcorner = split(/\ /, $bblores);
  $bbleft = $bbcorner[0];
  $bbbottom = $bbcorner[1];
  $bbright = $bbcorner[2];
  $bbtop = $bbcorner[3];
  $xtotal = ($bbright + $bbleft) * $sf;
  $ytotal = ($bbtop + $bbbottom) * $sf;
  if (($xtotal > 0) && ($ytotal > 0)) {
    $molps = $sf . " " . $sf . " scale\n" . $molps;  ## insert the PS "scale" command
    #print "low res: $bblores  .... max X: $bbright, max Y: $bbtop \n";
    if ($verbose > 1) {
      print "$filename  $xtotal x $ytotal pt\n";
    }
  } else {
    $xtotal = 99;
    $ytotal = 55;
    $molps = "%!PS-Adobe
    /Helvetica findfont 14 scalefont setfont
    10 30 moveto
    (2D structure) show
    10 15 moveto
    (not available) show
    showpage\n";
    if ($verbose > 1) {
      print "writing empty file\n";
    }
  }	
  # fix a little color problem with mol2ps v0.4
  $molps =~ s/gsave/gsave\ 0\ 0\ 0\ setrgbcolor/g;
  $gsopt1 = " -r300 -dGraphicsAlphaBits=4 -dTextAlphaBits=4 -dDEVICEWIDTHPOINTS=";
  $gsopt1 = $gsopt1 . $xtotal . " -dDEVICEHEIGHTPOINTS=" . $ytotal;
  $gsopt1 = $gsopt1 . " -sOutputFile=" . $filename;
  $gscmd = "$GHOSTSCRIPT -q -sDEVICE=$gsdevice -dNOPAUSE -dBATCH " . $gsopt1 . " - ";
  if ($ostype eq 2) {
    $dummy = filterthroughcmd2($molps, $gscmd);
  } else {
    system("echo \"$molps\" \| $gscmd");
  }
  #  system("echo \"$molps\" \| $gscmd");
}

sub pic2d_check() {
  if ($verbose > 0) { 
    print " checking for $pic2dtable in data collection $dbnum (type = $dbtype)\n";
  }
  $createcmd = "CREATE TABLE IF NOT EXISTS $pic2dtable (
  `$idname` INT(11) NOT NULL DEFAULT '0',
  `type` TINYINT NOT NULL DEFAULT '1' COMMENT '1 = png',
  `status` TINYINT NOT NULL DEFAULT '0' COMMENT '0 = does not exist, 1 = OK, 2 = OK, but do not show, 3 = to be created/updated, 4 = to be deleted',
  PRIMARY KEY ($idname)
  ) ENGINE=MYISAM CHARACTER SET $mysql_charset COLLATE $mysql_collation COMMENT='Housekeeping for 2D depiction'";
  #print "$createcmd\n";
  $dbh->do($createcmd);
  # next, check if number of records in structable and pic2dtable matches
  $sth0 = $dbh->prepare("SELECT COUNT($idname) AS itemcount FROM $structable");
  $sth0->execute();
  while ($ref0 = $sth0->fetchrow_hashref()) {
    $scount = $ref0->{'itemcount'};
  }
  $sth0->finish();
  $s_nchunks = int( (($scount + 99) / 100) );
  $sth0 = $dbh->prepare("SELECT COUNT($idname) AS itemcount FROM $pic2dtable");
  $sth0->execute();
  while ($ref0 = $sth0->fetchrow_hashref()) {
    $pcount = $ref0->{'itemcount'};
  }
  $sth0->finish();
  $p_nchunks = int( (($scount + 99) / 100) );
  #print "structures: $scount  pic2d entries: $pcount\n";
  if ($scount != $pcount) {
    $last_id = 0;
    for ($j = 0; $j < $s_nchunks; $j++) {
      $sth2 = $dbh->prepare("SELECT $idname FROM $structable WHERE $idname > $last_id LIMIT 0, 100");
      $sth2->execute();
      while ($ref2 = $sth2->fetchrow_hashref()) {
        $my_id   = $ref2->{"$idname"};
        $last_id = $my_id;
        $sth3 = $dbh->prepare("SELECT COUNT($idname) AS mycount FROM $pic2dtable WHERE $idname = $my_id");
        $sth3->execute();
        $my_count = 1;
        while ($ref3 = $sth3->fetchrow_hashref()) {
          $my_count   = $ref3->{"mycount"};
        }
        $sth3->finish();
        if ($my_count == 0) {
          # now add the missing record
          if ($verbose > 1) { print "   adding missing record ($my_id)\n"; }
          $updstr = "INSERT INTO $pic2dtable VALUES ($my_id, \"1\", \"3\" )";
          $dbh->do($updstr);
        }
      }  # while ..
      $sth2->finish();
    }   # for ..
  }
}


sub filterthroughcmd {
  $input   = shift;
  $cmd     = shift;
  open(FHSUB, "echo \"$input\"|$cmd 2>&1 |");   # stderr must be redirected to stdout
  $res      = "";                               # because the Ghostscript "bbox" device
  while($line = <FHSUB>) {                      # writes to stderr
    $res = $res . $line;
  }
  return $res;
}

sub filterthroughcmd2 {                         # workaround for Windows 
  $input   = shift;
  $cmd     = shift;
  ($tmpfh, $tmpfilename) = tempfile(UNLINK => 1);
  $input =~ s/\\\$/\$/g;
  $input =~ s/\r//g;
  $input =~ s/\n/\r\n/g;
  print $tmpfh "$input\n";
  #open(FHSUB, "type $tmpfilename |$cmd 2>&1 |");   # stderr must be redirected to stdout
  open(FHSUB, "$cmd < $tmpfilename 2>&1 |");   # stderr must be redirected to stdout
  $res      = "";                               # because the Ghostscript "bbox" device
  while($line = <FHSUB>) {                      # writes to stderr
    $res = $res . $line;
  }
  close $tmpfh;
  return $res;
}

sub ltrim() {
  $subline1 = shift;
  $subline1 =~ s/^\ +//g;
  return $subline1;
}

sub getostype() {
  $os  = "";
  $osresult = 1;
  $os  = uc($ENV{OS});
  if ($os eq "") { $os = uc($ENV{OSTYPE}); }
  if (index($os,"WINDOWS")>=0) {
    $osresult = 2;
  }
  return $osresult;
}

sub getval() {
  my $argline = shift;
  chomp($argline);
  my @argpart = split(/=/, $argline);
  my $argval = "";
  if ($#argpart > 0) { $argval = $argpart[1]; }
  return $argval;
}
